# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2017 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging as python_logging
import re
import time
import traceback
import uuid

from oslo_config import cfg
from oslo_log import log as logging
import pbr.version
import pecan
from pecan import request
import six

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.vpn import manager as vpn_manager


LOG = logging.getLogger(__name__)

LOOKING_GLASS_BASE = "looking-glass"


def expose(*args, **kwargs):
    """Helper function so we don't have to specify json for everything."""
    kwargs.setdefault('content_type', 'application/json')
    kwargs.setdefault('template', 'json')
    return pecan.expose(*args, **kwargs)


def when(index, *args, **kwargs):
    """Helper function so we don't have to specify json for everything."""
    kwargs.setdefault('content_type', 'application/json')
    kwargs.setdefault('template', 'json')
    return index.when(*args, **kwargs)


class PingController(object):

    def __init__(self):
        # Random generated sequence number
        self.sequence = int(uuid.uuid4())

    @expose(generic=True)
    def index(self):
        return self.sequence


class VPNManagerController(object):

    def __init__(self):
        self.manager = vpn_manager.VPNManager.get_instance()

    @staticmethod
    def stop():
        vpn_manager.VPNManager.get_instance().stop()


def check_attach_parameters(params, attach):
    LOG.debug("checking params: %s", params)
    paramList = ('vpn_instance_id', 'mac_address', 'ip_address',
                 'local_port')
    if attach:
        paramList += ('vpn_type', 'import_rt', 'export_rt', 'gateway_ip')

    for paramName in paramList:
        if paramName not in params:
            LOG.warning("Mandatory parameter '%s' is missing", paramName)
            pecan.abort(400,
                        "Mandatory parameter '%s' is missing" % paramName)

    # if local_port is not a dict, then assume it designates a linux
    # interface
    if (isinstance(params['local_port'], six.string_types) or
            isinstance(params['local_port'], six.text_type)):
        params['local_port'] = {'linuxif': params['local_port']}

    # if import_rt or export_rt are strings, convert them into lists
    for param in ('import_rt', 'export_rt'):
        if (isinstance(params[param], six.string_types) or
                isinstance(params[param], six.text_type)):
            try:
                params[param] = re.split(',+ *', params[param])
            except Exception:
                pecan.abort(400, "Unable to parse string into a list: '%s'" %
                            params[param])

    if not ('linuxif' in params['local_port'] or
            'evpn' in params['local_port']):
        pecan.abort(400, "Mandatory key is missing in local_port parameter"
                    "(linuxif, or evpn)")

    if not isinstance(params.get('advertise_subnet', False), bool):
        pecan.abort(400, "'advertise_subnet' must be a boolean")

    return params


class AttachController(VPNManagerController):

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    @when(index, method='POST')
    def process(self):
        try:
            attach_params = request.json
        except Exception:
            LOG.error('attach_localport: No local port details received')
            pecan.abort(400, 'No local port details received')

        attach_params = check_attach_parameters(attach_params, True)

        try:
            LOG.debug('Local port attach received: %s', attach_params)

            self.manager.plug_vif_to_vpn(
                attach_params['vpn_instance_id'],
                attach_params['vpn_type'],
                attach_params['import_rt'],
                attach_params['export_rt'],
                attach_params['mac_address'],
                attach_params['ip_address'],
                attach_params['gateway_ip'],
                attach_params['local_port'],
                attach_params.get('linuxbr'),
                attach_params.get('advertise_subnet', False),
                attach_params.get('readvertise'),
                attach_params.get('attract_traffic'),
                attach_params.get('lb_consistent_hash_order', 0),
                attach_params.get('fallback'),
            )
        except exc.APIException as e:
            LOG.warning('attach_localport: API parameter error: %s', e)
            pecan.abort(400, "API parameter error: %s" % e)
        except Exception as e:
            LOG.error('attach_localport: An error occurred during local port'
                      ' plug to VPN: %s', e)
            LOG.info(traceback.format_exc())
            pecan.abort(500, 'An error occurred during local port plug to VPN')


class DetachController(VPNManagerController):

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    @when(index, method='POST')
    def process(self):
        try:
            detach_params = request.json
        except Exception:
            LOG.error('detach_localport: No local port details received')
            pecan.abort(400, 'No local port details received')

        detach_params = check_attach_parameters(detach_params, attach=False)

        try:
            LOG.debug('Local port detach received: %s', detach_params)
            self.manager.unplug_vif_from_vpn(
                detach_params['vpn_instance_id'],
                detach_params['mac_address'],
                detach_params['ip_address'],
                detach_params['local_port'],
                detach_params.get('advertise_subnet', False)
            )
        except exc.APIException as e:
            LOG.warning('detach_localport: API parameter error: %s', e)
            pecan.abort(400, "API parameter error: %s" % e)
        except Exception as e:
            LOG.error('detach_localport: An error occurred during local port'
                      ' unplug from VPN: %s', e)
            LOG.info(traceback.format_exc())
            pecan.abort(500, 'An error occurred during local port unplug from '
                        'VPN')


class LookingGlassController(VPNManagerController,
                             lg.LookingGlassMixin):

    def __init__(self):
        super(LookingGlassController, self).__init__()

        self.start_time = time.time()

        lg.set_references_root(LOOKING_GLASS_BASE)
        lg.set_reference_path("BGP_WORKERS", ["bgp", "workers"])
        lg.set_reference_path("VPN_INSTANCES", ["vpns", "instances"])
        lg.set_reference_path("DATAPLANE_DRIVERS",
                              ["vpns", "dataplane", "drivers"])

        self.catchall_lg_log_handler = lg.LookingGlassLogHandler()
        python_logging.getLogger().addHandler(self.catchall_lg_log_handler)

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='GET')
    def process(self, *url_path_elements):

        path_prefix = "%s://%s/%s" % (
            request.scheme,  # http
            request.host,
            LOOKING_GLASS_BASE,
        )

        try:
            lg_info = self.get_looking_glass_info(path_prefix,
                                                  url_path_elements)
            if lg_info is None:
                raise lg.NoSuchLookingGlassObject(path_prefix,
                                                  url_path_elements[0])
            return lg_info
        except lg.NoSuchLookingGlassObject as e:
            LOG.info('looking_glass: %s', repr(e))
            pecan.abort(404, repr(e))
        except Exception as e:
            LOG.error('looking_glass: An error occurred: %s', e)
            LOG.error(traceback.format_exc())
            pecan.abort(500, 'Server error')

    @when(index, method='DELETE')
    @when(index, method='POST')
    @when(index, method='PUT')
    def not_supported(self):
        pecan.abort(405)

    # Looking glass hooks #################

    def get_lg_map(self):
        return {
            "summary":  (lg.SUBITEM, self.get_lg_summary),
            "config":   (lg.SUBTREE, self.get_lg_config),
            "bgp":      (lg.DELEGATE, self.manager.bgp_manager),
            "vpns":     (lg.DELEGATE, self.manager),
            "logs":     (lg.SUBTREE, self.get_logs)
        }

    def get_lg_config(self, path_prefix):
        return {section: utils.osloconfig_json_serialize(cfg.CONF[section])
                for section in ('COMMON', 'API', 'BGP',
                                'DATAPLANE_DRIVER_IPVPN',
                                'DATAPLANE_DRIVER_EVPN')
                }

    def get_lg_summary(self):
        return {
            "BGP_established_peers":
                self.manager.bgp_manager.get_established_peers_count(),
            "local_routes_count":
                self.manager.bgp_manager.rtm.
                get_local_routes_count(),
            "received_routes_count":
                self.manager.bgp_manager.rtm.
                get_received_routes_count(),
            "vpn_instances_count": self.manager.get_vpn_instances_count(),
            "warnings_and_errors": len(self.catchall_lg_log_handler),
            "start_time": time.strftime("%Y-%m-%d %H:%M:%S",
                                        time.localtime(self.start_time)),
            "version":  (pbr.version.VersionInfo('networking-bagpipe')
                         .release_string())
        }

    def get_logs(self, path_prefix):
        return [{'level': record.levelname,
                 'time':
                 self.catchall_lg_log_handler.formatter.formatTime(record),
                 'name': record.name,
                 'message': record.msg}
                for record in self.catchall_lg_log_handler.get_records()]


class RootController(object):

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='POST')
    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    ping = PingController()
    attach_localport = AttachController()
    detach_localport = DetachController()

    def stop(self):
        VPNManagerController.stop()


# there is a '-' in the LOOKING_GLASS_BASE name, so we have to use pecan.route
pecan.route(RootController, LOOKING_GLASS_BASE, LookingGlassController())



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import six

import socket

import httplib2
import json

from copy import deepcopy

from collections import defaultdict
from collections import namedtuple
from collections import OrderedDict

import netaddr

from oslo_config import cfg

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from oslo_concurrency import lockutils

from oslo_service import loopingcall

from networking_bagpipe._i18n import _

from networking_bagpipe.rpc import agent as bagpipe_agent_rpc
from networking_bagpipe.rpc.client import topics_BAGPIPE

from networking_bagpipe.agent.bgpvpn import rpc_agent as bgpvpn_agent_rpc
from networking_bagpipe.agent.bgpvpn.rpc_client import topics_BAGPIPE_BGPVPN

from neutron.agent.common import ovs_lib
from neutron.callbacks import events
from neutron.callbacks import registry
from neutron.callbacks import resources
from neutron.common import topics
from neutron.conf.agent import common as config

from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc

from neutron.conf.plugins.ml2.drivers import ovs_conf

from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants\
    as a_const
from neutron.plugins.ml2.drivers.openvswitch.agent import ovs_neutron_agent
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

LOG = logging.getLogger(__name__)

DEFAULT_GATEWAY_MAC = "00:00:5e:00:43:64"
FALLBACK_SRC_MAC = "00:00:5e:2a:10:00"

BAGPIPE_NOTIFIER = 'bagpipe'
BGPVPN_NOTIFIER = 'bgpvpn'
BAGPIPE_NOTIFIERS = [BAGPIPE_NOTIFIER, BGPVPN_NOTIFIER]

EVPN = 'evpn'
IPVPN = 'ipvpn'
VPN_TYPES = [EVPN, IPVPN]

BGPVPN_L2 = 'l2vpn'
BGPVPN_L3 = 'l3vpn'
BGPVPN_TYPES = [BGPVPN_L2, BGPVPN_L3]
BGPVPN_TYPES_MAP = {BGPVPN_L2: EVPN, BGPVPN_L3: IPVPN}

LINUXIF_PREFIX = "patch2tun"

bagpipe_bgp_opts = [
    cfg.IntOpt('ping_interval', default=10,
               help=_("The number of seconds the bagpipe-bgp client will "
                      "wait between polling for restart detection.")),
    cfg.StrOpt('bagpipe_bgp_ip', default='127.0.0.1',
               help=_("bagpipe-bgp REST service IP address.")),
    cfg.IntOpt('bagpipe_bgp_port', default=8082,
               help=_("bagpipe-bgp REST service IP port.")),
    cfg.StrOpt('mpls_bridge', default='br-mpls',
               help=_("OVS MPLS bridge to use")),
    cfg.StrOpt('tun_to_mpls_peer_patch_port', default='patch-to-mpls',
               help=_("OVS Peer patch port in tunnel bridge to MPLS bridge "
                      "(traffic to MPLS bridge)")),
    cfg.StrOpt('tun_from_mpls_peer_patch_port', default='patch-from-mpls',
               help=_("OVS Peer patch port in tunnel bridge to MPLS bridge "
                      "(traffic from MPLS bridge)")),
    cfg.StrOpt('mpls_to_tun_peer_patch_port', default='patch-to-tun',
               help=_("OVS Peer patch port in MPLS bridge to tunnel bridge "
                      "(traffic to tunnel bridge)")),
    cfg.StrOpt('mpls_from_tun_peer_patch_port', default='patch-from-tun',
               help=_("OVS Peer patch port in MPLS bridge to tunnel bridge "
                      "(traffic from tunnel bridge)")),
    cfg.StrOpt('mpls_to_int_peer_patch_port', default='patch-mpls-to-int',
               help=_("OVS Peer patch port in MPLS bridge to int bridge "
                      "(traffic to int bridge)")),
    cfg.StrOpt('int_from_mpls_peer_patch_port', default='patch-int-from-mpls',
               help=_("OVS Peer patch port in int bridge to MPLS bridge "
                      "(traffic from MPLS bridge)")),
]

cfg.CONF.register_opts(bagpipe_bgp_opts, "BAGPIPE")
ovs_conf.register_ovs_agent_opts()
config.register_agent_state_opts_helper(cfg.CONF)


class BGPAttachmentNotFound(n_exc.NotFound):
    message = "Local port %(local_port)s details could not be found"


class BaGPipeBGPException(n_exc.NeutronException):
    message = "An exception occurred when calling bagpipe-bgp \
               REST service: %(reason)s"


class HTTPClientBase(object):
    """An HTTP client base class"""

    def __init__(self, host="127.0.0.1", port=8082,
                 client_name="HTTP client base"):
        """Create a new HTTP client

        :param host: HTTP server IP address
        :param port: HTTP server port
        """
        self.host = host
        self.port = port
        self.client_name = client_name

    def do_request(self, method, action, body=None):
        LOG.debug("bagpipe-bgp client request: %s %s [%s]" %
                  (method, action, str(body)))

        if isinstance(body, dict):
            body = json.dumps(body)
        try:
            headers = {'User-Agent': self.client_name,
                       "Content-Type": "application/json",
                       "Accept": "application/json"}
            uri = "http://%s:%s/%s" % (self.host, self.port, action)

            http = httplib2.Http()
            response, content = http.request(uri, method, body, headers)
            LOG.debug("bagpipe-bgp returns [%s:%s]" %
                      (str(response.status), content))

            if response.status == 200:
                if content and len(content) > 1:
                    return json.loads(content)
            else:
                reason = (
                    "An HTTP operation has failed on bagpipe-bgp."
                )
                raise BaGPipeBGPException(reason=reason)
        except (socket.error, IOError) as e:
            reason = "Failed to connect to bagpipe-bgp: %s" % str(e)
            raise BaGPipeBGPException(reason=reason)

    def get(self, action):
        return self.do_request("GET", action)

    def post(self, action, body=None):
        return self.do_request("POST", action, body=body)

    def put(self, action, body=None):
        return self.do_request("PUT", action, body=body)

    def delete(self, action):
        return self.do_request("DELETE", action)


class DummyOVSAgent(ovs_neutron_agent.OVSNeutronAgent):
    # this class is used only to 'borrow' setup_entry_for_arp_reply
    # from OVSNeutronAgent
    arp_responder_enabled = True

    def __init__(self):
        pass


GatewayInfo = namedtuple('GatewayInfo', ['mac', 'ip'])
NO_GW_INFO = GatewayInfo(None, None)


def default_gw_info():
    return GatewayInfo(mac=None, ip=None)


def has_attachement(bgpvpn_info, vpn_type):
    return (bgpvpn_info[vpn_type].get('import_rt') or
            bgpvpn_info[vpn_type].get('export_rt'))


def has_ipvpn_attachement(bgpvpn_info):
    return has_attachement(bgpvpn_info, IPVPN)


class BaGPipeBGPAgent(HTTPClientBase,
                      bagpipe_agent_rpc.BaGPipeAgentRpcCallBackMixin,
                      bgpvpn_agent_rpc.BGPVPNAgentRpcCallBackMixin
                      ):
    """Implements a BaGPipe-BGP REST client"""

    # bagpipe-bgp status
    BAGPIPEBGP_UP = 'UP'
    BAGPIPEBGP_DOWN = 'DOWN'

    def __init__(self, agent_type, connection, int_br=None, tun_br=None):

        """Create a new BaGPipe-BGP REST service client.

        :param agent_type: bagpipe-bgp agent type (Linux bridge or OVS)
        :param connection: RPC Connection
        :param int_br: OVS integration bridge
        :param tun_br: OVS tunnel bridge
        """
        super(BaGPipeBGPAgent,
              self).__init__(cfg.CONF.BAGPIPE.bagpipe_bgp_ip,
                             cfg.CONF.BAGPIPE.bagpipe_bgp_port, agent_type)

        self.agent_type = agent_type

        self.ping_interval = cfg.CONF.BAGPIPE.ping_interval

        self.reg_attachments = defaultdict(list)
        self.gateway_info_for_net = defaultdict(default_gw_info)

        self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN
        self.seq_num = 0

        # OVS-specific setup
        if self.agent_type == n_const.AGENT_TYPE_OVS:
            self.int_br = int_br
            self.tun_br = tun_br
            self.setup_mpls_br()

            registry.subscribe(self.ovs_restarted,
                               resources.AGENT,
                               events.OVS_RESTARTED)

        # RPC setup
        if self.agent_type == n_const.AGENT_TYPE_LINUXBRIDGE:
            connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                             topics_BAGPIPE,
                                                             topics.UPDATE,
                                                             cfg.CONF.host),
                                       [self], fanout=False)
        else:
            LOG.info("bagpipe-l2 RPCs disabled for OVS bridge")

        connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                         topics_BAGPIPE_BGPVPN,
                                                         topics.UPDATE),
                                   [self], fanout=True)
        connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                         topics_BAGPIPE_BGPVPN,
                                                         topics.UPDATE,
                                                         cfg.CONF.host),
                                   [self], fanout=False)

        # Starts a greenthread for bagpipe-bgp status polling
        self._start_bagpipe_bgp_status_polling(self.ping_interval)

        self.vlan_manager = vlanmanager.LocalVlanManager()

    def _check_bagpipe_bgp_status(self):
        """Trigger refresh on bagpipe-bgp restarts

        Check if bagpipe-bgp has restarted while sending ping request
        to detect sequence number change.
        If a restart is detected, re-send all registered attachments to
        bagpipe-bgp.
        """
        new_seq_num = self.request_ping()

        # Check bagpipe-bgp restart
        if new_seq_num != self.seq_num:
            if new_seq_num != -1:
                if self.seq_num != 0:
                    LOG.warning("bagpipe-bgp restart detected...")
                else:
                    LOG.info("bagpipe-bgp successfully detected")

                self.seq_num = new_seq_num
                self.bagpipe_bgp_status = self.BAGPIPEBGP_UP

                # Re-send all registered attachments to bagpipe-bgp
                if self.reg_attachments:
                    LOG.info("Sending all registered attachments to "
                             "bagpipe-bgp")
                    LOG.debug("Registered attachments list: %s" %
                              self.reg_attachments)
                    for __, attachment_list in six.iteritems(
                            self.reg_attachments):
                        if attachment_list:
                            for attachment in attachment_list:
                                self._do_local_port_plug(attachment)
                else:
                    LOG.info("No attachment to send to bagpipe-bgp")
            else:
                self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN

    def _start_bagpipe_bgp_status_polling(self, ping_interval=10):
        # Start bagpipe-bgp status polling at regular interval
        status_loop = loopingcall.FixedIntervalLoopingCall(
            self._check_bagpipe_bgp_status)
        status_loop.start(interval=ping_interval,
                          initial_delay=ping_interval)  # TM: why not zero ?

    def _get_reg_attachment_for_port(self, network_id, local_port_id):
        # Retrieve local port index and attachement details in BGP
        # registered attachments list for the specified network and port
        # identifiers

        LOG.debug("Getting registered attachment for port %s on network %s" %
                  (local_port_id, network_id))

        index = -1
        details = None
        for i, attachment in enumerate(self.reg_attachments[network_id]):
            if (attachment['port_id'] == local_port_id):
                LOG.debug("Registered attachment found at index %s: %s" %
                          (i, attachment))
                index = i
                details = attachment
                break

        if index == -1:
            LOG.info("No registered attachment found for port %s",
                     local_port_id)
            raise BGPAttachmentNotFound(local_port=local_port_id)

        return index, details

    def _copy_local_port_common_details(self, local_port_details):
        local_port_copy = {}
        for key in ['local_port']:
            local_port_copy[key] = local_port_details[key]

        for optional in ['linuxbr']:
            if optional in local_port_details:
                local_port_copy[optional] = local_port_details[optional]

        return local_port_copy

    def _gen_fallback_info(self, dst_mac):
        return {'fallback': {'dst_mac': dst_mac,
                             'src_mac': FALLBACK_SRC_MAC,
                             'ovs_port_number': self.patch_mpls_to_int_ofport
                             }
                }

    def _add_fallback_info(self, source_details, details, vpn_type):
        # if a gateway_mac is known, we will add fallback information
        # to 'details'

        if (self.agent_type == n_const.AGENT_TYPE_OVS and
                vpn_type == IPVPN and source_details.get('gateway_mac')):
            details.update(
                self._gen_fallback_info(source_details.get('gateway_mac'))
            )

    def _copy_notifier_vpn_details(self, notifier_details, vpn_type):
        notifier_copy = {}
        for key in ['mac_address', 'ip_address', 'gateway_ip',
                    'import_rt', 'export_rt']:
            notifier_copy[key] = notifier_details[key]

        for optional in ['readvertise', 'attract_traffic']:
            if optional in notifier_details:
                notifier_copy[optional] = notifier_details[optional]

        self._add_fallback_info(notifier_details, notifier_copy, vpn_type)

        return notifier_copy

    def _populate_vpntype_ipaddress2details(self, vpn_type, notifier,
                                            local_port_details,
                                            vpntype_ipaddress2details):
        # (used only as a helper for _get_local_port_plug_details)
        # will populate vpntype_ipaddress2details dict

        if (notifier in local_port_details and
                vpn_type in local_port_details[notifier] and
                has_attachement(local_port_details[notifier], vpn_type)):
            notifier_details = local_port_details[notifier][vpn_type]
            LOG.debug("Notifier %s details: %s" % (vpn_type,
                                                   notifier_details))

            if ((vpn_type, notifier_details['ip_address']) in
                    vpntype_ipaddress2details):
                plug_details = vpntype_ipaddress2details[
                    (vpn_type, notifier_details['ip_address'])
                    ]

                plug_details['import_rt'] += notifier_details['import_rt']
                plug_details['export_rt'] += notifier_details['export_rt']

                vpntype_ipaddress2details[
                    (vpn_type, notifier_details['ip_address'])
                    ].update(plug_details)
            else:
                plug_details = {
                    'vpn_type': vpn_type,
                    'vpn_instance_id': '%s_%s' %
                    (local_port_details['vpn_instance_id'], vpn_type),
                }

                plug_details.update(
                    self._copy_local_port_common_details(local_port_details))

                if IPVPN == vpn_type:
                    for vpntype_ipaddress in vpntype_ipaddress2details:
                        if EVPN in vpntype_ipaddress:
                            plug_details['local_port'] = {EVPN: {
                                'id': '%s_evpn' %
                                local_port_details['vpn_instance_id']}
                                }

                plug_details.update(
                    self._copy_notifier_vpn_details(notifier_details, vpn_type)
                )

                vpntype_ipaddress2details[(vpn_type,
                                           notifier_details['ip_address'])
                                          ] = deepcopy(plug_details)

            if 'static_routes' in notifier_details:
                for static_route in notifier_details['static_routes']:
                    plug_details['ip_address'] = static_route
                    plug_details.update({'advertise_subnet': True})
                    vpntype_ipaddress2details[
                        (vpn_type, static_route)] = deepcopy(plug_details)

    def _get_local_port_plug_details(self, local_port_details, vpn_types=None,
                                     notifiers=None):
        # Construct plug details list for specified notifiers corresponding to
        # the given port

        all_plug_details = {}
        vpntype_ipaddress2details = OrderedDict()

        plug_types = vpn_types if vpn_types else VPN_TYPES
        plug_notifiers = notifiers if notifiers else BAGPIPE_NOTIFIERS

        # NOTE(tmorin): could use itertools.product
        for (vpn_type, notifier) in [
                (x, y) for x in plug_types for y in plug_notifiers]:
            self._populate_vpntype_ipaddress2details(vpn_type,
                                                     notifier,
                                                     local_port_details,
                                                     vpntype_ipaddress2details)

        LOG.debug('Local port (VPN type, IP address) map details %s' %
                  vpntype_ipaddress2details)

        for (vpntype_ipaddress, plug_details) in six.iteritems(
                vpntype_ipaddress2details):
            if vpntype_ipaddress[0] in all_plug_details:
                all_plug_details[vpntype_ipaddress[0]].append(plug_details)
            else:
                all_plug_details[vpntype_ipaddress[0]] = list([plug_details])

        LOG.debug('Local port %s details %s' % (local_port_details['port_id'],
                                                all_plug_details))

        return all_plug_details

    # bagpipe-bgp REST API requests
    # ----------------------------------------
    def request_ping(self):
        """Send ping request to bagpipe-bgp to get sequence number"""
        try:
            response = self.get('ping')
            LOG.debug("bagpipe-bgp PING response received with "
                      "sequence number %s" % response)
            return response
        except BaGPipeBGPException as e:
            LOG.warning(str(e))
            return -1

    def send_attach_local_port(self, local_port_details):
        """Send local port attach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('attach_localport', local_port_details)
                LOG.debug("Local port has been attached to bagpipe-bgp with "
                          "details %s" % local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't attach local port on bagpipe-bgp: %s", str(e))
        else:
            LOG.debug("Local port not yet attached to bagpipe-bgp (not up)")

    def send_detach_local_port(self, local_port_details):
        """Send local port detach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('detach_localport', local_port_details)
                LOG.debug("Local port has been detached from bagpipe-bgp "
                          "with details %s" % local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach local port from bagpipe-bgp: %s",
                          str(e))
                raise
        else:
            LOG.debug("Local port not yet detached from bagpipe-bgp (not up)")

    def setup_mpls_br(self):
        '''Setup the MPLS bridge for bagpipe-bgp.

        Creates MPLS bridge, and links it to the integration and tunnel
        bridges using patch ports.

        :param mpls_br: the name of the MPLS bridge.
        '''
        mpls_br = cfg.CONF.BAGPIPE.mpls_bridge
        self.mpls_br = ovs_lib.OVSBridge(mpls_br)

        if not self.mpls_br.bridge_exists(mpls_br):
            LOG.error("Unable to enable MPLS on this agent, MPLS bridge "
                      "%(mpls_br)s doesn't exist. Agent terminated!",
                      {"mpls_br": mpls_br})
            exit(1)

        # patch ports for traffic from tun bridge to mpls bridge
        self.patch_tun_to_mpls_ofport = self.tun_br.add_patch_port(
            cfg.CONF.BAGPIPE.tun_to_mpls_peer_patch_port,
            cfg.CONF.BAGPIPE.mpls_from_tun_peer_patch_port)
        self.patch_mpls_from_tun_ofport = self.mpls_br.add_patch_port(
            cfg.CONF.BAGPIPE.mpls_from_tun_peer_patch_port,
            cfg.CONF.BAGPIPE.tun_to_mpls_peer_patch_port)

        # patch ports for traffic from mpls bridge to tun bridge
        self.patch_mpls_to_tun_ofport = self.mpls_br.add_patch_port(
            cfg.CONF.BAGPIPE.mpls_to_tun_peer_patch_port,
            cfg.CONF.BAGPIPE.tun_from_mpls_peer_patch_port)
        self.patch_tun_from_mpls_ofport = self.tun_br.add_patch_port(
            cfg.CONF.BAGPIPE.tun_from_mpls_peer_patch_port,
            cfg.CONF.BAGPIPE.mpls_to_tun_peer_patch_port)

        # patch ports for traffic from mpls bridge to int bridge
        self.patch_mpls_to_int_ofport = self.mpls_br.add_patch_port(
            cfg.CONF.BAGPIPE.mpls_to_int_peer_patch_port,
            cfg.CONF.BAGPIPE.int_from_mpls_peer_patch_port)
        self.patch_int_from_mpls_ofport = self.int_br.add_patch_port(
            cfg.CONF.BAGPIPE.int_from_mpls_peer_patch_port,
            cfg.CONF.BAGPIPE.mpls_to_int_peer_patch_port)

        if (int(self.patch_tun_to_mpls_ofport) < 0 or
                int(self.patch_mpls_from_tun_ofport) < 0 or
                int(self.patch_mpls_to_tun_ofport) < 0 or
                int(self.patch_tun_from_mpls_ofport) < 0 or
                int(self.patch_int_from_mpls_ofport) < 0 or
                int(self.patch_mpls_to_int_ofport) < 0):
            LOG.error("Failed to create OVS patch port. Cannot have "
                      "MPLS enabled on this agent, since this version "
                      "of OVS does not support patch ports. "
                      "Agent terminated!")
            exit(1)

        patch_int_ofport = self.tun_br.get_port_ofport(
            cfg.CONF.OVS.tun_peer_patch_port)

        # In br-tun, redirect all traffic from VMs toward a BGPVPN
        # default gateway MAC address to the MPLS bridge.
        #
        # (priority >0 is needed or we hit the rule redirecting unicast to
        # the UCAST_TO_TUN table)
        self.tun_br.add_flow(
            table=a_const.PATCH_LV_TO_TUN,
            priority=1,
            in_port=patch_int_ofport,
            dl_dst=DEFAULT_GATEWAY_MAC,
            actions="output:%s" % self.patch_tun_to_mpls_ofport
        )

        # Redirect traffic from the MPLS bridge to br-int
        self.tun_br.add_flow(in_port=self.patch_tun_from_mpls_ofport,
                             actions="output:%s" % patch_int_ofport)

    def ovs_restarted(self, resources, event, trigger):
        self.setup_mpls_br()
        self.ovs_restarted_bgpvpn()
        # TODO(tmorin): need to handle restart on bagpipe-bgp side, in the
        # meantime after an OVS restart, restarting bagpipe-bgp is required

    def _get_local_port_for_attach(self, port_id, net_id):
        if self.agent_type == n_const.AGENT_TYPE_LINUXBRIDGE:
            port_name = LinuxBridgeManager.get_tap_device_name(port_id)
            bridge_name = LinuxBridgeManager.get_bridge_name(net_id)

            return {
                'linuxbr': bridge_name,
                'local_port': {
                    'linuxif': port_name
                }
            }
        elif self.agent_type == n_const.AGENT_TYPE_OVS:
            vlan = self.vlan_manager.get(net_id).vlan

            return {
                'local_port': {
                    'linuxif': "%s:%s" % (LINUXIF_PREFIX, vlan),
                    'ovs': {
                        'plugged': True,
                        'port_number': self.patch_mpls_from_tun_ofport,
                        'to_vm_port_number': self.patch_mpls_to_tun_ofport,
                        'vlan': vlan
                    }
                }
            }

    def _copy_port_network_details(self, port_details, vpn_type):
        network_details = {}
        for key in ['ip_address', 'mac_address', 'gateway_ip', 'gateway_mac']:
            network_details[key] = port_details.get(key)

        self._add_fallback_info(port_details, network_details, vpn_type)

        return network_details

    def _format_port_details(self, port_details, notifier):
        """Format port details as follows:

        {
            mac_address: <PORT_MAC>,
            ip_address: <PORT_IP>,
            gateway_ip: <PORT_GATEWAY>,
            gateway_mac: <GW_MAC>,
            evpn | ipvpn: {
                import_rts: [<ASN:NN>],
                export_rts: [<ASN:NN>]
            }
        }

        to

        {
            evpn | ipvpn: {
                mac_address: <PORT_MAC>,
                ip_address: <PORT_IP>,
                gateway_ip: <PORT_GATEWAY>,
                fallback: {...}
                import_rts: [<ASN:NN>],
                export_rts: [<ASN:NN>]
            }
        }
        """
        copy_details = deepcopy(port_details)
        formatted_details = {}
        for vpn_type in VPN_TYPES:
            if vpn_type in copy_details:
                formatted_details[vpn_type] = copy_details.pop(vpn_type)
                formatted_details[vpn_type].update(copy_details)

            elif notifier == BGPVPN_NOTIFIER:
                formatted_details[vpn_type] = (
                    self._copy_port_network_details(copy_details, vpn_type)
                )

        return formatted_details

    @lockutils.synchronized('bagpipe-bgp-agent')
    def _add_local_port_details(self, notifier, network_id, port_id,
                                port_info):
        """Create/update local port details in registered attachments list

        As follows:
        {
            port_id: <PORT_UUID>,
            vpn_instance_id: <NETWORK_UUID|PORT_UUID>,
            local_port: {
                linuxif: <PORT_NAME>,

                # Optional parameter only used in OVS agent
                ovs: {
                    plugged: True,
                    port_number: <PATCH_MPLS_FROM_TUN_OFPORT>,
                    to_vm_port_number: <PATCH_MPLS_TO_TUN_OFPORT>,
                    vlan: <LOCAL_VLAN_IDENTIFIER>
                }
            }

            # Optional parameter only used in Linux Bridge agent
            linuxbr: <LINUX_BRIDGE_NAME>,

            bagpipe: {
                evpn | ipvpn: {
                    mac_address: <PORT_MAC>,
                    ip_address: <PORT_IP>,
                    gateway_ip: <PORT_GATEWAY>,
                    import_rts: [<ASN:NN>, ...],
                    export_rts: [<ASN:NN>, ...]
                    ...
                }
            },
            bgpvpn: {
                evpn | ipvpn: {
                    mac_address: <PORT_MAC>,
                    ip_address: <PORT_IP>,
                    gateway_ip: <PORT_GATEWAY>,
                    import_rts: [<ASN:NN>, ...],
                    export_rts: [<ASN:NN>, ...]
                    ...
                }
            }
        }

        and return formatted port details to notifier for plug/unplug action.
        """
        try:
            # Update local port details to registered attachments list
            index, port_details = self._get_reg_attachment_for_port(network_id,
                                                                    port_id)

            # Format port details from notified informations
            formatted_details = self._format_port_details(port_info, notifier)

            if notifier in port_details:
                port_details[notifier].update(formatted_details)
            else:
                port_details[notifier] = formatted_details

            self.reg_attachments[network_id][index] = port_details
        except BGPAttachmentNotFound:
            LOG.info("Adding local port to attachments list")
            # Add local port details to registered attachments list
            port_details = {'port_id': port_id}

            # Update with bagpipe-bgp local_port details depending on
            # agent type
            port_details.update(self._get_local_port_for_attach(port_id,
                                                                network_id))

            if 'vpn_instance_id' not in port_info:
                port_details['vpn_instance_id'] = network_id
            else:
                port_details['vpn_instance_id'] = (
                    port_info.pop('vpn_instance_id'))

            # Format port details from notified informations
            port_details[notifier] = self._format_port_details(port_info,
                                                               notifier)
            self.reg_attachments[network_id].append(port_details)
        finally:
            LOG.debug("Updated attachments list: %s" %
                      self.reg_attachments[network_id])

        return port_details

    @lockutils.synchronized('bagpipe-bgp-agent')
    def _remove_registered_attachment(self, net_id, index,
                                      notifier, delete=True):
        """Remove/update a registered attachment for a notifier

        Remove/update notifier registered attachment at this index
        for the specified network, only if no
        exception occurred on bagpipe-bgp.
        """
        if (notifier in self.reg_attachments[net_id][index]):
            if notifier == BGPVPN_NOTIFIER and not delete:
                bgpvpn_reg_attachment = (
                    self.reg_attachments[net_id][index][notifier]
                )
                for vpn_type in VPN_TYPES:
                    if vpn_type in bgpvpn_reg_attachment:
                        for rt_type in ['import_rt', 'export_rt']:
                            if rt_type in bgpvpn_reg_attachment[vpn_type]:
                                del bgpvpn_reg_attachment[vpn_type][rt_type]
            elif delete:
                del self.reg_attachments[net_id][index][notifier]

            LOG.debug("Checking to remove attachment: %s",
                      self.reg_attachments[net_id][index])
            if not any(key in self.reg_attachments[net_id][index]
                       for key in BAGPIPE_NOTIFIERS):
                LOG.debug("Deleting attachment...")
                del self.reg_attachments[net_id][index]

                # Check if removing last registered attachment
                if not self.reg_attachments[net_id]:
                    LOG.debug("No attachment on network, deleting...")
                    del self.reg_attachments[net_id]

    @log_helpers.log_method_call
    def _enable_gw_redirect(self, vlan, gateway_ip):
        # Add ARP responder entry for default gateway in br-tun

        # We may compete with the ARP responder entry for the real MAC
        # if the router is on a network node and we are a compute node,
        # so we must add our rule with a higher priority

        # TODO(tmorin): Ideally we should use install_arp_responder, after
        # extending it to accept a 'priority' parameter, but we'd need to make
        # sure that delete_arp_responder (including the OF native
        # implementation) delete only rule with its cookie

        # (mostly copy-pasted ovs_ofctl....install_arp_responder)
        actions = a_const.ARP_RESPONDER_ACTIONS % {
            'mac': netaddr.EUI(DEFAULT_GATEWAY_MAC, dialect=netaddr.mac_unix),
            'ip': netaddr.IPAddress(gateway_ip),
        }
        self.tun_br.add_flow(table=a_const.ARP_RESPONDER,
                             priority=2,  # see above
                             dl_vlan=vlan,
                             proto='arp',
                             arp_op=0x01,  # the rule in install_arp_responder
                                           # does not restrict to requests,
                                           # but it seems what is right to do
                             arp_tpa='%s' % gateway_ip,
                             actions=actions)

    @log_helpers.log_method_call
    def _disable_gw_redirect(self, vlan, gateway_ip):
        # Remove ARP responder entry for default gateway in br-tun
        self.tun_br.delete_flows(
            table=a_const.ARP_RESPONDER,
            dl_vlan=vlan,
            proto='arp',
            arp_op=0x01,
            arp_tpa='%s' % gateway_ip)

        # TODO(tmorin): nothing currently prevents the base agent from
        # deleting our rule on Router unplug (because its delete_flows is
        # not restricted to deleting rules having its own cookie)

    @log_helpers.log_method_call
    def _hide_real_gw_arp(self, vlan, gateway_info):
        # Kill ARP replies for the gateway IP coming on br-int from the real
        # router, if any.
        #
        # NOTE(tmorin): we assume that the router MAC exists only in this vlan.
        # Doing filtering based on the local vlan would be better, but
        # we can't do this in br-int because this bridge does tagging based
        # on ovs-vsctl port tags.
        self.int_br.add_flow(table=a_const.LOCAL_SWITCHING,
                             priority=2,
                             proto='arp',
                             arp_op=0x2,
                             dl_src=gateway_info.mac,
                             arp_sha=gateway_info.mac,
                             arp_spa=gateway_info.ip,
                             actions="drop")

        # ARP requests from the real gateway need to
        # have their IP address changed to hide the gateway
        # address or the VMs will use it to update their
        # ARP cache implicitly. Below we overwrite it with 0.0.0.0.
        self.int_br.add_flow(table=a_const.LOCAL_SWITCHING,
                             priority=2,
                             proto='arp',
                             arp_op=0x01,
                             dl_src=gateway_info.mac,
                             arp_spa=gateway_info.ip,
                             arp_sha=gateway_info.mac,
                             actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL")

    @log_helpers.log_method_call
    def _unhide_real_gw_arp(self, vlan, gateway_info):
        LOG.debug("unblocking ARP from real gateway for vlan %d (%s)",
                  vlan, gateway_info)
        self.int_br.delete_flows(table=a_const.LOCAL_SWITCHING,
                                 proto='arp',
                                 dl_src=gateway_info.mac,
                                 arp_sha=gateway_info.mac)

    def _do_local_port_plug(self, local_port_details):
        """Send local port attach request to bagpipe-bgp."""
        all_plug_details = (
            self._get_local_port_plug_details(local_port_details)
        )

        # First plug E-VPNs because they could be plugged into IP-VPNs
        for vpn_type in [t for t in VPN_TYPES
                         if t in all_plug_details]:
            for plug_detail in all_plug_details[vpn_type]:
                self.send_attach_local_port(plug_detail)

    def _do_local_port_unplug(self, local_port_details, vpn_types=None,
                              notifiers=None):
        """Send local port detach request to bagpipe-bgp."""
        all_unplug_details = (
            self._get_local_port_plug_details(local_port_details,
                                              vpn_types=vpn_types,
                                              notifiers=notifiers)
        )

        # First unplug IP-VPNs because E-VPNs could be plugged into them
        for vpn_type in [t for t in VPN_TYPES[::-1]
                         if t in all_unplug_details]:
            for unplug_detail in all_unplug_details[vpn_type]:
                self.send_detach_local_port(unplug_detail)

    @log_helpers.log_method_call
    def _check_arp_voodoo_plug(self, net_id, gateway_info):

        if (self.agent_type != n_const.AGENT_TYPE_OVS):
            return

        # See if we need to update gateway redirection and gateway ARP
        # voodoo

        vlan = self.vlan_manager.get(net_id).vlan

        # NOTE(tmorin): can be improved, only needed on first plug...
        self._enable_gw_redirect(vlan, gateway_info.ip)

        # update real gateway ARP blocking...
        # remove old ARP blocking ?
        if (net_id in self.gateway_info_for_net and
                self.gateway_info_for_net.get(net_id).mac is not None):
            self._unhide_real_gw_arp(vlan,
                                     self.gateway_info_for_net.get(net_id))
        # add new ARP blocking ?
        if gateway_info.mac:
            self._hide_real_gw_arp(vlan, gateway_info)

    @log_helpers.log_method_call
    def _check_arp_voodoo_unplug(self, net_id):

        if (self.agent_type != n_const.AGENT_TYPE_OVS):
            return

        # if last port for this network, then cleanup gateway redirection
        # NOTE(tmorin): shouldn't we check for last *ipvpn* attachment?
        if len(self.reg_attachments.get(net_id)) == 1:
            LOG.debug("last unplug, undoing voodoo ARP")
            # NOTE(tmorin): vlan lookup might break if port is already
            # unplugged from bridge ?
            vlan = self.vlan_manager.get(net_id).vlan
            self._disable_gw_redirect(vlan,
                                      self.gateway_info_for_net[net_id].ip)
            if self.gateway_info_for_net.get(net_id).mac is not None:
                self._unhide_real_gw_arp(vlan,
                                         self.gateway_info_for_net[net_id])

    # BaGPipe RPC callbacks
    # ----------------------
    def bagpipe_port_attach(self, context, port_bagpipe_info):
        LOG.debug("bagpipe_port_attach received with port info: %s",
                  port_bagpipe_info)
        port_id = port_bagpipe_info.pop('id')
        net_id = port_bagpipe_info.pop('network_id')

        # Add/Update local port details in registered attachments list
        port_details = self._add_local_port_details(BAGPIPE_NOTIFIER,
                                                    net_id,
                                                    port_id,
                                                    port_bagpipe_info)

        # Attach port on bagpipe-bgp
        LOG.debug(
            "Attaching BaGPipe L2 port %s on bagpipe-bgp with "
            "details %s", port_details['local_port']['linuxif'], port_details)

        self._do_local_port_plug(port_details)

    def bagpipe_port_detach(self, context, port_bagpipe_info):
        LOG.debug("bagpipe_port_detach received with port info %s",
                  port_bagpipe_info)
        port_id = port_bagpipe_info['id']
        net_id = port_bagpipe_info['network_id']

        # Detach port from bagpipe-bgp
        LOG.debug("Detaching port %s from bagpipe-bgp", port_id)
        try:
            index, port_details = self._get_reg_attachment_for_port(net_id,
                                                                    port_id)
        except BGPAttachmentNotFound as e:
            LOG.error("bagpipe-bgp agent inconsistent for BaGPipe L2 or "
                      "updated with another detach: %s", str(e))
        else:
            try:
                self._do_local_port_unplug(port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach port from bagpipe-bgp: %s", str(e))
            finally:
                self._remove_registered_attachment(net_id, index,
                                                   BAGPIPE_NOTIFIER)

    # BGPVPN callbacks
    # -----------------------------
    def _update_local_port_bgpvpn_details(self, network_id, index,
                                          evpn, ipvpn):
        attachment = self.reg_attachments[network_id][index]

        if evpn:
            if EVPN in attachment[BGPVPN_NOTIFIER]:
                attachment[BGPVPN_NOTIFIER][EVPN].update(evpn)

        if ipvpn:
            if IPVPN in attachment[BGPVPN_NOTIFIER]:
                attachment[BGPVPN_NOTIFIER][IPVPN].update(ipvpn)

        self.reg_attachments[network_id][index] = attachment

        return attachment

    @log_helpers.log_method_call
    def _attach_all_ports_on_bgpvpn(self, network_id, bgpvpn):
        # Attach all ports on BaGPipe-BGP for to the specified BGP VPN

        evpn = bgpvpn.get(EVPN, bgpvpn.get(BGPVPN_L2))
        ipvpn = bgpvpn.get(IPVPN, bgpvpn.get(BGPVPN_L3))

        new_gw_info = GatewayInfo(
            bgpvpn.get('gateway_mac'),
            self.gateway_info_for_net[network_id].ip
        )

        for index, __ in enumerate(self.reg_attachments[network_id]):
            updated_attachment = (
                self._update_local_port_bgpvpn_details(network_id, index,
                                                       evpn, ipvpn)
            )

            if has_ipvpn_attachement(updated_attachment.get(BGPVPN_NOTIFIER)):
                if new_gw_info != NO_GW_INFO:
                    # TODO(tmorin): this does not need to be done more than
                    # one time for a given network:
                    self._check_arp_voodoo_plug(
                        network_id,
                        new_gw_info
                    )
                if new_gw_info.mac:
                    updated_attachment.get(BGPVPN_NOTIFIER).get(IPVPN).update(
                        {'gateway_mac': new_gw_info.mac}
                    )

            self._do_local_port_plug(updated_attachment)

        if bgpvpn.get('gateway_mac'):
            self.gateway_info_for_net[network_id] = new_gw_info

    @log_helpers.log_method_call
    def _detach_all_ports_from_bgpvpn(self, network_id):
        # Detach all ports from BaGPipe-BGP for the specified BGP VPN

        for index, port_details in enumerate(self.reg_attachments[network_id]):
            try:
                self._do_local_port_unplug(port_details,
                                           notifiers=[BGPVPN_NOTIFIER])
            except BaGPipeBGPException as e:
                LOG.error("Can't detach port from bagpipe-bgp: %s", str(e))
            finally:
                if has_ipvpn_attachement(port_details.get(BGPVPN_NOTIFIER)):
                    # TODO(tmorin): this does not need to be done more than one
                    # time for a given network:
                    self._check_arp_voodoo_unplug(network_id)
                self._remove_registered_attachment(network_id, index,
                                                   BGPVPN_NOTIFIER,
                                                   delete=False)

    def ovs_restarted_bgpvpn(self):
        for (network_id, attachments) in self.reg_attachments.items():
            gw_info = self.gateway_info_for_net[network_id]

            if attachments and gw_info != NO_GW_INFO:
                for attachment in attachments:
                    if has_ipvpn_attachement(attachment.get(BGPVPN_NOTIFIER)):
                        self._check_arp_voodoo_plug(network_id, gw_info)
                        break

    @log_helpers.log_method_call
    def create_bgpvpn(self, context, bgpvpn):
        self.update_bgpvpn(context, bgpvpn)

    @log_helpers.log_method_call
    def update_bgpvpn(self, context, bgpvpn):
        self._attach_all_ports_on_bgpvpn(bgpvpn['network_id'], bgpvpn)

    @log_helpers.log_method_call
    def delete_bgpvpn(self, context, bgpvpn):
        self._detach_all_ports_from_bgpvpn(bgpvpn['network_id'])

    @log_helpers.log_method_call
    def bgpvpn_port_attach(self, context, port_bgpvpn_info):
        port_id = port_bgpvpn_info.pop('id')
        net_id = port_bgpvpn_info.pop('network_id')

        # Map l2vpn, l3vpn keys to evpn, ipvpn ones
        for bgpvpn_type in BGPVPN_TYPES:
            if bgpvpn_type in port_bgpvpn_info:
                port_bgpvpn_info[BGPVPN_TYPES_MAP[bgpvpn_type]] = (
                    port_bgpvpn_info.pop(bgpvpn_type)
                )

        gateway_info = GatewayInfo(port_bgpvpn_info.get('gateway_mac'),
                                   port_bgpvpn_info.get('gateway_ip'))

        if IPVPN in port_bgpvpn_info:
            self._check_arp_voodoo_plug(net_id, gateway_info)

        # Remember the gateway MAC
        self.gateway_info_for_net[net_id] = gateway_info

        # Add/Update port BGPVPN details in registered attachments list
        port_details = self._add_local_port_details(BGPVPN_NOTIFIER,
                                                    net_id,
                                                    port_id,
                                                    port_bgpvpn_info)

        # Attach port on bagpipe-bgp
        LOG.debug("Attaching BGPVPN port %s on bagpipe-bgp with details %s" %
                  (port_details['local_port']['linuxif'], port_details))

        self._do_local_port_plug(port_details)

    @log_helpers.log_method_call
    def bgpvpn_port_detach(self, context, port_bgpvpn_info):
        port_id = port_bgpvpn_info['id']
        net_id = port_bgpvpn_info['network_id']

        # Retrieve local port details index in BGP registered attachment list
        try:
            index, details = self._get_reg_attachment_for_port(net_id,
                                                               port_id)
        except BGPAttachmentNotFound as e:
            LOG.warning("BGPVPN port detach, but no previous attachment found:"
                        " %s", str(e))
        else:
            try:
                self._do_local_port_unplug(details)

                if has_ipvpn_attachement(details.get(BGPVPN_NOTIFIER)):
                    self._check_arp_voodoo_unplug(net_id)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach BGPVPN port from bagpipe-bgp %s",
                          str(e))
            finally:
                self._remove_registered_attachment(net_id, index,
                                                   BGPVPN_NOTIFIER)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_bagpipe.agent.bgpvpn import agent_extension as bagpipe_agt_ext

from neutron.agent.common import ovs_lib
from neutron.plugins.ml2.drivers.openvswitch.agent.common \
    import constants as ovs_agt_constants
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_agt
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent \
    import ovs_test_base

from neutron_lib import constants as const


class TestOVSAgentExtension(ovs_test_base.OVSOFCtlTestBase):

    def setUp(self):
        super(TestOVSAgentExtension, self).setUp()
        self.agent_ext = bagpipe_agt_ext.BagpipeBgpvpnAgentExtension()
        self.connection = mock.Mock()

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent.BaGPipeBGPAgent')
    def test_init(self, mocked_bagpipe_bgp_agent):
        int_br = self.br_int_cls("br-int")
        tun_br = self.br_tun_cls("br-tun")
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(int_br,
                                                               tun_br)

        self.agent_ext.consume_api(agent_extension_api)
        self.agent_ext.initialize(self.connection,
                                  ovs_agt_constants.EXTENSION_DRIVER_TYPE,
                                  )

        mocked_bagpipe_bgp_agent.assert_called_once_with(
            const.AGENT_TYPE_OVS,
            self.connection,
            int_br=mock.ANY,
            tun_br=mock.ANY
            )

        call_kwargs = mocked_bagpipe_bgp_agent.call_args_list[0][1]

        self.assertIsInstance(call_kwargs['int_br'], ovs_lib.OVSBridge)
        self.assertIsInstance(call_kwargs['tun_br'], ovs_lib.OVSBridge)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import socket
import traceback

from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import worker


keep_attributes_default = [exa.Attribute.CODE.NEXT_HOP,
                           exa.Attribute.CODE.PMSI_TUNNEL,
                           exa.Attribute.CODE.MED,
                           exa.Attribute.CODE.EXTENDED_COMMUNITY,  # FIXME
                           exa.Attribute.CODE.LOCAL_PREF]


class FilteredRouteEntry(engine.RouteEntry):

    def __init__(self, re, keep_attributes=None):
        if keep_attributes is None:
            keep_attributes = keep_attributes_default

        attributes = exa.Attributes()
        for (attribute_id, attribute) in re.attributes.items():
            if attribute_id in keep_attributes:
                attributes.add(attribute)

        engine.RouteEntry.__init__(self, re.nlri, None, attributes)


def filtered_routes(routes):
    return [FilteredRouteEntry(route) for route in routes]


# def _compare_routes(self, route_a, route_b):
#         """
#         should return:
#          - an int>0 if route_a is better than route_b
#          - an int<0 if route_b is better than route_a
#          - else 0
#         """

# TODO(tmorin): both comparison should first compare local_pref and MAC
# Mobility if present


def compare_no_ecmp(self, route_a, route_b):
    '''comparison operator without ECMP

    This compares the two routes in a consistent fashion, but two routes
    will never be considered of equal cost.
    The comparison is 'salted' so that two distinct VRFs (e.g. on two distinct
    bagpipe-bgp instances will not necessarily elect the same route as the
    best one.
    '''
    self.log.debug("compare_no_ecmp used")
    salt = socket.gethostname() + self.name
    hash_a = hash(salt + repr(route_a))
    hash_b = hash(salt + repr(route_b))
    cmp_hashes = (hash_a > hash_b) - (hash_b > hash_a)

    if cmp_hashes != 0:
        return cmp_hashes

    return (route_a > route_b) - (route_b > route_a)


def compare_ecmp(self, route_a, route_b):
    self.log.debug("compare_ecmp used")
    return 0


@six.add_metaclass(abc.ABCMeta)
class TrackerWorker(worker.Worker, lg.LookingGlassLocalLogger):

    def __init__(self, bgp_manager, worker_name,
                 compare_routes=compare_no_ecmp):

        worker.Worker.__init__(self, bgp_manager, worker_name)
        lg.LookingGlassLocalLogger.__init__(self)

        # dict: entry -> list of routes:
        self.tracked_entry_2_routes = dict()
        # dict: entry -> set of best_routes:
        self.tracked_entry_2_best_routes = dict()

        self._compare_routes = compare_routes

    @log_decorator.log
    def _on_event(self, event):
        new_route = event.route_entry
        filtered_new_route = FilteredRouteEntry(new_route)

        entry = self._route_2_tracked_entry(new_route)

        if entry is None:
            self.log.debug("Route not mapped to a tracked entry, ignoring: %s",
                           new_route)
            return

        self.log.debug("tracked_entry for this route: %s (type: %s)",
                       TrackerWorker._display_entry(entry), type(entry))

        self._dump_state()

        all_routes = self.tracked_entry_2_routes.setdefault(entry, [])

        self.log.debug("We currently have %d route%s for this entry",
                       len(all_routes), utils.plural(all_routes))

        if event.type == engine.RouteEvent.ADVERTISE:

            withdrawn_best_routes = []

            best_routes = self.tracked_entry_2_best_routes.get(entry)

            if best_routes is None:
                self.log.debug("We had no route for this entry (%s)")
                self.tracked_entry_2_best_routes[entry] = set([new_route])
                best_routes = set()
                self.log.debug("Calling new_best_route")
                self._call_new_best_route(entry, filtered_new_route)
            else:
                if event.replaced_route is not None:
                    self.log.debug("Will remove replaced route from all_routes"
                                   " and best_routes: %s",
                                   event.replaced_route)
                    try:
                        all_routes.remove(event.replaced_route)
                    except ValueError:
                        # we did not have any route for this entry
                        self.log.error("replaced_route is an entry for which "
                                       "we had no route ??? (bug ?)")

                    if event.replaced_route in best_routes:
                        self.log.debug(
                            "Removing replaced_route from best_routes")
                        best_routes.remove(event.replaced_route)
                        withdrawn_best_routes.append(event.replaced_route)
                    else:
                        self.log.debug("replaced_route is not in best_routes")
                        self.log.debug("best_routes: %s", best_routes)
                else:
                    self.log.debug("No replaced route to remove")

                call_new_best_route_4_all = False
                if len(best_routes) == 0:
                    self.log.debug("All best routes have been replaced")
                    self._recompute_best_routes(all_routes, best_routes)
                    if best_routes:
                        current_best = six.next(iter(best_routes))
                        self.log.debug("We'll need to call new_best_route for "
                                       "all our new best routes")
                        call_new_best_route_4_all = True
                    else:
                        current_best = None
                        call_new_best_route_4_all = False
                else:
                    # (if there is more than one route in the best routes, we
                    # take the first one)
                    current_best = six.next(iter(best_routes))

                    self.log.debug("Current best route: %s", current_best)

                    if new_route == current_best:
                        self.log.info("New route is a route we already had, "
                                      "nothing to do.")
                        # nothing to do
                        return

                # let's find if we need to update our best routes
                if current_best:
                    route_comparison = self._compare_routes(self, new_route,
                                                            current_best)
                else:
                    route_comparison = 1

                self.log.debug("route_comparison: %d", route_comparison)

                if route_comparison > 0:
                    # new_route is a strictly better route than any current
                    # one, discard all the current best routes
                    self.log.debug("Replacing all best routes with new one")
                    withdrawn_best_routes.extend(best_routes.copy())
                    best_routes.clear()
                    best_routes.add(new_route)
                    self._call_new_best_route(entry, filtered_new_route)
                    call_new_best_route_4_all = False
                elif route_comparison == 0:
                    # new_route is as good as the current ones
                    self.log.debug("Adding new_route to best_routes...")

                    if call_new_best_route_4_all:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)

                    # We'll do a call to self._new_best_route... *only* if the
                    # new_route is different from all current best routes. This
                    # comparison uses FilteredRouteEntry to *not* take into
                    # account .source (the BGP peer which advertized the route)
                    # and only takes into account a specific set of BGP
                    # attributes.
                    # TODO(tmorin): explain more on theses BGP attributes
                    #   related to the cases where a route is re-advertized
                    #   with updated attributes
                    is_really_new = (FilteredRouteEntry(new_route) not in
                                     filtered_routes(best_routes))

                    best_routes.add(new_route)

                    if is_really_new:
                        self.log.debug("Calling self._new_best_route since we "
                                       "yet had no such route in best routes")
                        self._call_new_best_route(entry, filtered_new_route)
                    else:
                        self.log.debug("Not calling _new_best_route since we "
                                       "had received a similar route already")
                else:
                    self.log.debug("The route is no better than current "
                                   "best ones")

                    if call_new_best_route_4_all:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)

            # We need to call self._best_route_removed for routes that where
            # implicitly withdrawn, but only if they don't have an equal route
            # (in the sense of FilteredRouteEntry) in best_routes
            filtered_best_routes = filtered_routes(best_routes)
            self.log.debug("Considering implicitly withdrawn best routes")
            for r in withdrawn_best_routes:
                filtered_r = FilteredRouteEntry(r)
                if filtered_r not in filtered_best_routes:
                    self.log.debug("   calling self._best_route_removed for "
                                   "route: %s (not last)", filtered_r)
                    self._call_best_route_removed(entry,
                                                  filtered_r,
                                                  last=False)
                else:
                    self.log.debug("   not calling self._best_route_removed "
                                   "for route: %s", filtered_r)

            # add the route to the list of routes for this entry
            self.log.debug("Adding route to all_routes for this entry")
            all_routes.append(new_route)

        else:  # RouteEvent.WITHDRAW

            withdrawn_route = new_route

            self.log.debug("Removing route from all_routes for this entry")

            # let's update known routes for this entry
            try:
                all_routes.remove(withdrawn_route)
            except ValueError:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we"
                               " had no route ??? (not supposed to happen)")

            # let's now update best routes
            best_routes = self.tracked_entry_2_best_routes.get(entry)

            if best_routes is None:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we "
                               "had no route: not supposed to happen!")
                return

            if withdrawn_route in best_routes:
                self.log.debug("The event received is about a route which"
                               " is among the best routes for this entry")
                # remove the route from best_routes
                best_routes.remove(withdrawn_route)

                withdrawn_route_is_last = False
                if len(best_routes) == 0:
                    # we don't have any best route left...
                    self._recompute_best_routes(all_routes, best_routes)

                    if len(best_routes) > 0:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)
                    else:
                        self.log.debug("Cleanup all_routes and best_routes")
                        withdrawn_route_is_last = True
                        del self.tracked_entry_2_best_routes[entry]
                        del self.tracked_entry_2_routes[entry]

                self.log.debug("Calling best_route_removed...?")
                # We need to call self._best_route_removed, but only if the
                # withdrawn route does not have an equal route in
                # best_routes (in the sense of FilteredRouteEntry)
                filtered_withdrawn_route = FilteredRouteEntry(withdrawn_route)
                if (filtered_withdrawn_route not
                        in filtered_routes(best_routes)):
                    self.log.debug("Calling best_route_removed: %s(last:%s)",
                                   filtered_withdrawn_route,
                                   withdrawn_route_is_last)
                    self._call_best_route_removed(entry,
                                                  filtered_withdrawn_route,
                                                  withdrawn_route_is_last)
                else:
                    self.log.debug("No need to call bestRouteRemved: %s",
                                   filtered_withdrawn_route)

            else:
                self.log.debug("The event received is not related to any "
                               "of the best routes for this entry")
                # no need to update our best route list
                pass

        self.log.info("We now have %d route%s for this entry.",
                      len(all_routes), utils.plural(all_routes))

        self._dump_state()

    def _recompute_best_routes(self, all_routes, best_routes):
        '''internal method to update best routes

        update best_routes to contain the best routes from all_routes, based
        on _compare_routes
        '''

        new_best_routes = []
        for route in all_routes:
            if len(new_best_routes) == 0:
                new_best_routes = [route]
                continue

            comparison = self._compare_routes(self, route, new_best_routes[0])
            if comparison > 0:
                new_best_routes = [route]
            elif comparison == 0:
                new_best_routes.append(route)

        best_routes.clear()
        best_routes.update(new_best_routes)

        self.log.debug("Recomputed new best routes: %s", best_routes)

    def _call_new_best_route_for_routes(self, entry, routes):
        self.log.debug("Calling new_best_route for routes, without dups")
        self.log.debug("   Routes: %s", routes)
        routes_no_dups = set([FilteredRouteEntry(r) for r in routes])
        self.log.debug("   After filtering duplicates: %s", routes_no_dups)
        for route in routes_no_dups:
            self._call_new_best_route(entry, route)

    def _call_new_best_route(self, entry, new_route):
        try:
            self._new_best_route(entry, new_route)
        except Exception as e:
            self.log.error("Exception in <subclass>._new_best_route: %s", e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.info("%s", traceback.format_exc())

    def _call_best_route_removed(self, entry, old_route, last):
        try:
            self._best_route_removed(entry, old_route, last)
        except Exception as e:
            self.log.error("Exception in <subclass>._best_route_removed: %s",
                           e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.info("%s", traceback.format_exc())

    # Callbacks for subclasses ########################

    @abc.abstractmethod
    def _route_2_tracked_entry(self, route):
        """Hook to control mapping from a route to a tracked entry

        This method is how the subclass maps a BGP route into an object that
        the TrackerWorker code will track.

        For instance, a VPN VRF is expected to keep track of IP prefixes;
        hence the route2tracked_entry code for a VRF could return the IP prefix
        in the VPNv4 route.
        The result will be that the TrackerWorker code will keep track, for a
        each prefix, of all the routes and of the best routes.
        """
        pass

    @abc.abstractmethod
    def _new_best_route(self, entry, new_route):
        '''New Best Route hook

        A new best route has been advertized for this tracked entry
        '''
        pass

    @abc.abstractmethod
    def _best_route_removed(self, entry, old_route, last):
        '''Best Route removed hook

        A route that was a best route for this tracked entry has been
        removed. The 'last' flag indicates if this was the last route
        for this tracked entry.
        '''
        pass

    # Debug support methods #########

    def _dump_state(self):
        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug("--- tracked_entry_2_routes ---")
            for (entry, routes) in six.iteritems(self.tracked_entry_2_routes):
                self.log.debug(
                    "  Entry: %s", TrackerWorker._display_entry(entry))
                for route in routes:
                    self.log.debug("    Route: %s", route)

            self.log.debug("--- tracked_entry_2_best_routes ---")
            for (entry, best_routes) in \
                    six.iteritems(self.tracked_entry_2_best_routes):
                self.log.debug("  Entry: %s",
                               TrackerWorker._display_entry(entry))
                for route in best_routes:
                    self.log.debug("    Route: %s", route)

            self.log.debug("--- ---")

    @staticmethod
    def _display_entry(entry):
        if (isinstance(entry, tuple) and len(entry) > 0 and
                (isinstance(entry[0], type) or
                 isinstance(entry[0], six.class_types))):
            return repr(tuple([entry[0].__name__] + list(entry[1:])))
        else:
            return repr(entry)

    # Looking glass ###########

    def get_lg_map(self):
        return {"received_routes": (lg.SUBTREE, self.get_lg_all_routes),
                "best_routes": (lg.SUBTREE, self.get_lg_best_routes)}

    def get_lg_all_routes(self, path_prefix):
        return self._get_lg_routes(path_prefix,
                                   self.tracked_entry_2_routes)

    def get_lg_best_routes(self, path_prefix):
        return self._get_lg_routes(path_prefix,
                                   self.tracked_entry_2_best_routes)

    def _get_lg_routes(self, path_prefix, route_dict):
        routes = {}
        for entry in six.iterkeys(route_dict):
            entry_repr = self._display_entry(entry)
            routes[entry_repr] = [route.get_looking_glass_info(path_prefix)
                                  for route in route_dict[entry]]
        return routes



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import threading

from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.engine import exa


LOG = logging.getLogger(__name__)

MAX_RD_LOCAL_ID = 2**16-1


class RDAllocator(lg.LookingGlassMixin):

    def __init__(self, prefix):
        self.prefix = prefix
        self.current_id = random.randint(100, 200)
        self.rds = dict()

        self.lock = threading.Lock()

    @utils.synchronized
    def get_new_rd(self, description):

        if self.current_id == MAX_RD_LOCAL_ID+1:
            LOG.error("All the %d possible local ids have been used at least "
                      "once, and this piece of code doesn't know how to reuse "
                      "them", MAX_RD_LOCAL_ID)
            raise Exception("Out of local ids")

        rd = exa.RouteDistinguisher.fromElements(self.prefix, self.current_id)
        self.current_id += 1
        self.rds[rd] = description

        LOG.debug("Allocated route distinguisher %s for '%s'", rd, description)
        return rd

    @utils.synchronized
    def release(self, rd):
        if rd in self.rds:
            LOG.debug("Released route distinguisher %s ('%s')",
                      rd, self.rds[rd])
            del self.rds[rd]
        else:
            LOG.warn("Asked to release a non registered route distinguisher: "
                     "%s", rd)

    def get_lg_local_info(self, prefix):
        return self.rds



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from oslo_config import cfg
from oslo_log import log

from neutron.agent import securitygroups_rpc
from neutron.db import api as db_api
from neutron.db import models_v2

from neutron.plugins.common import constants as p_constants

from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const

from networking_bagpipe.driver.type_route_target import TYPE_ROUTE_TARGET
from networking_bagpipe.rpc import client as bagpipe_rpc_client

from sqlalchemy.orm import exc


LOG = log.getLogger(__name__)

ml2_bagpipe_opts = [
    cfg.IntOpt('as_number', default=64512,
               help="Autonomous System number"),
]

cfg.CONF.register_opts(ml2_bagpipe_opts, "ml2_bagpipe")


class NoNetworkInfoForPort(Exception):
    pass


def get_network_info_for_port(session, port_id):
    """Get Port network informations from DB

    Get network informations (MAC, IP and gateway addresses and
    subnet mask) from database associated to a port
    """
    LOG.debug("get_network_info_for_port() called for port %s" % port_id)

    with session.begin(subtransactions=True):
        try:
            net_info = (session.
                        query(models_v2.Port.mac_address,
                              models_v2.IPAllocation.ip_address,
                              models_v2.Subnet.cidr,
                              models_v2.Subnet.gateway_ip).
                        join(models_v2.IPAllocation).
                        join(models_v2.Subnet,
                             models_v2.IPAllocation.subnet_id ==
                             models_v2.Subnet.id).
                        filter(models_v2.Subnet.ip_version == 4).
                        filter(models_v2.Port.id == port_id).one())
            return net_info
        except exc.NoResultFound:
            raise NoNetworkInfoForPort(port_id)


class BaGPipeMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """ML2 Mechanism driver for bagpipe-bgp

    This mechanism driver uses RPCs toward compute node agents to trigger
    the attachment of VM ports in E-VPN VPN instances.
    """

    def __init__(self):
        sg_enabled = securitygroups_rpc.is_firewall_enabled()
        super(BaGPipeMechanismDriver, self).__init__(
            n_const.AGENT_TYPE_LINUXBRIDGE,
            portbindings.VIF_TYPE_BRIDGE,
            {portbindings.CAP_PORT_FILTER: sg_enabled})
        self.agent_notify = bagpipe_rpc_client.BaGPipeAgentNotifyAPI()

    def initialize(self):
        self.rpc_ctx = n_context.get_admin_context_without_session()
        self.migrated_ports = {}

    def get_allowed_network_types(self, agent):
        return (agent['configurations'].get('tunnel_types', []) +
                [p_constants.TYPE_LOCAL, p_constants.TYPE_FLAT,
                 p_constants.TYPE_VLAN])

    def get_mappings(self, agent):
        return agent['configurations'].get('interface_mappings', {})

    def _get_route_target(self, segment):
        if (segment and segment[api.NETWORK_TYPE] == TYPE_ROUTE_TARGET):
            return "%s:%s" % (cfg.CONF.ml2_bagpipe.as_number,
                              segment.get(api.SEGMENTATION_ID))
        else:
            LOG.warning("no RT for segment %s", segment)

    def check_segment_for_agent(self, segment, agent):
        mappings = agent['configurations'].get('bridge_mappings', {})
        tunnel_types = agent['configurations'].get('tunnel_types', [])
        LOG.debug("Checking segment: %(segment)s "
                  "for mappings: %(mappings)s "
                  "with tunnel_types: %(tunnel_types)s",
                  {'segment': segment, 'mappings': mappings,
                   'tunnel_types': tunnel_types})
        return (segment[api.NETWORK_TYPE] == TYPE_ROUTE_TARGET)

    def _get_network_info_for_port(self, port_id):
        """Get MAC, IP and Gw IP addresses informations for a specific port"""
        session = db_api.get_reader_session()
        (mac_address, ip_address, cidr, gateway_ip) = (
            get_network_info_for_port(session, port_id)
        )

        return {'mac_address': mac_address,
                'ip_address': ip_address + cidr[cidr.index('/'):],
                'gateway_ip': gateway_ip}

    def _retrieve_bagpipe_net_info_for_port(self, port_id, segment):
        """Retrieve BaGPipe network informations for a specific port

        {
            'network_id': <UUID>,
            'mac_address': '00:00:de:ad:be:ef',
            'ip_address': '10.0.0.2',
            'gateway_ip': '10.0.0.1',
            'evpn' : {
                'import_rt': ['12345:1', '12345:2', '12345:3'],
                'export_rt': ['12345:1', '12345:2', '12345:4']
            }
        }
        """
        bagpipe_network_info = {}

        # Check if port is connected on a BaGPipe network
        bagpipe_rt = self._get_route_target(segment)
        if bagpipe_rt:
            bagpipe_network_info.update(
                {'evpn': {
                    'import_rt': bagpipe_rt,
                    'export_rt': bagpipe_rt}}
            )
        else:
            LOG.debug("No E-VPN RT info for port %s", port_id)

        LOG.debug("Getting port %s network details", port_id)
        bagpipe_network_info.update(self._get_network_info_for_port(port_id))

        return bagpipe_network_info

    def delete_port_postcommit(self, context):
        port = context.current
        agent_host = context.host

        port_bagpipe_info = {'id': port['id'],
                             'network_id': port['network_id']}

        self.agent_notify.detach_port_from_bagpipe_network(
            self.rpc_ctx,
            port_bagpipe_info, agent_host
        )

    def update_port_postcommit(self, context):
        port = context.current
        orig = context.original
        agent_host = context.host

        if (context.host != context.original_host and
                context.status == n_const.PORT_STATUS_ACTIVE and
                not self.migrated_ports.get(orig['id'])):
            # Port has been migrated. We need to store the original
            # binding to send bagpipe informations for port to the
            # appropriate host
            self.migrated_ports[orig['id']] = (
                (orig, context.original_host))
        elif context.status != context.original_status:
            port_bagpipe_info = {'id': port['id'],
                                 'network_id': port['network_id']}

            if context.status == n_const.PORT_STATUS_ACTIVE:
                segment = context.bottom_bound_segment
                if not segment:
                    LOG.debug(("Port %(port)s updated by agent %(agent)s "
                               "isn't bound to any segment"),
                              {'port': port['id'], 'agent': agent_host})
                    return

                try:
                    bagpipe_network_info = (
                        self._retrieve_bagpipe_net_info_for_port(port['id'],
                                                                 segment)
                    )
                    port_bagpipe_info.update(bagpipe_network_info)
                    self.agent_notify.attach_port_on_bagpipe_network(
                        self.rpc_ctx,
                        port_bagpipe_info, agent_host
                    )
                except NoNetworkInfoForPort:
                    LOG.warning("No network info for port %s (v6 only?),"
                                " not attached!", port['id'])
            elif context.status == n_const.PORT_STATUS_DOWN:
                self.agent_notify.detach_port_from_bagpipe_network(
                    self.rpc_ctx,
                    port_bagpipe_info,
                    agent_host
                )
            elif context.status == n_const.PORT_STATUS_BUILD:
                orig = self.migrated_ports.pop(port['id'], None)
                if orig:
                    # this port has been migrated: Detach from BaGPipe
                    # network ? automatically handled by BGP ?
                    original_port = orig[0]
                    original_host = orig[1]
                    port_bagpipe_info.update({'id': original_port['id']})
                    self.agent_notify.detach_port_from_bagpipe_network(
                        self.rpc_ctx,
                        port_bagpipe_info,
                        original_host
                    )



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import constants
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import flowspec
from networking_bagpipe.bagpipe_bgp.engine import ipvpn as ipvpn_routes
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance


class DummyDataplaneDriver(dp_drivers.DummyDataplaneDriver):
    type = constants.IPVPN


class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
    # component managing a VRF:
    # - calling a driver to instantiate the dataplane
    # - registering to receive routes for the needed route targets
    # - calling the driver to setup/update/remove routes in the dataplane
    # - cleanup: calling the driver, unregistering for BGP routes

    type = constants.IPVPN
    afi = exa.AFI(exa.AFI.ipv4)
    safi = exa.SAFI(exa.SAFI.mpls_vpn)

    @log_decorator.log
    def __init__(self, *args, **kwargs):
        vpn_instance.VPNInstance.__init__(self, *args, **kwargs)
        self.readvertised = set()

    def _nlri_from(self, prefix, label, rd):
        assert rd is not None

        return ipvpn_routes.IPVPNRouteFactory(
            self.afi, prefix, label, rd,
            self.dp_driver.get_local_address())

    def generate_vif_bgp_route(self, mac_address, ip_prefix, plen, label, rd):
        # Generate BGP route and advertise it...
        nlri = self._nlri_from("%s/%s" % (ip_prefix, plen), label, rd)

        return engine.RouteEntry(nlri)

    def _get_local_labels(self):
        for port_data in self.mac_2_localport_data.values():
            yield port_data['label']

    def _imported(self, route):
        return len(set(route.route_targets).intersection(set(self.import_rts))
                   ) > 0

    def _to_readvertise(self, route):
        # Only re-advertise IP VPN routes (e.g. not Flowspec routes)
        if not isinstance(route.nlri, ipvpn_routes.IPVPN):
            return False

        rt_records = route.ecoms(exa.RTRecord)
        self.log.debug("RTRecords: %s (readvertise_to_rts:%s)",
                       rt_records,
                       self.readvertise_to_rts)

        readvertise_targets_as_records = [exa.RTRecord.from_rt(rt)
                                          for rt in self.readvertise_to_rts]

        if self.attract_traffic:
            readvertise_targets_as_records += [exa.RTRecord.from_rt(rt)
                                               for rt in self.attract_rts]

        if set(readvertise_targets_as_records).intersection(set(rt_records)):
            self.log.debug("not to re-advertise because one of the readvertise"
                           " or attract-redirect RTs is in RTRecords: %s",
                           set(readvertise_targets_as_records)
                           .intersection(set(rt_records)))
            return False

        return len(set(route.route_targets).intersection(
            set(self.readvertise_from_rts)
            )) > 0

    def _route_for_readvertisement(self, route, label, rd,
                                   lb_consistent_hash_order,
                                   do_default=False):
        prefix = "0.0.0.0/0" if do_default else route.nlri.cidr.prefix()

        nlri = self._nlri_from(prefix, label, rd)

        attributes = exa.Attributes()

        # new RTRecord = original RTRecord (if any) + orig RTs converted into
        # RTRecord attributes
        orig_rtrecords = route.ecoms(exa.RTRecord)
        rts = route.ecoms(exa.RTExtCom)
        add_rtrecords = [exa.RTRecord.from_rt(rt) for rt in rts]

        final_rtrecords = list(set(orig_rtrecords) | set(add_rtrecords))

        ecoms = self._gen_encap_extended_communities()
        ecoms.communities += final_rtrecords
        ecoms.communities.append(
            exa.ConsistentHashSortOrder(lb_consistent_hash_order))
        attributes.add(ecoms)

        entry = engine.RouteEntry(nlri, self.readvertise_to_rts, attributes)
        self.log.debug("RouteEntry for (re-)advertisement: %s", entry)
        return entry

    @log_decorator.log
    def _route_for_redirect_prefix(self, prefix):
        prefix_classifier = utils.dict_camelcase_to_underscore(
            self.attract_classifier)
        prefix_classifier['destination_prefix'] = prefix

        traffic_classifier = vpn_instance.TrafficClassifier(
            **prefix_classifier)
        self.log.debug("Advertising prefix %s for redirection based on "
                       "traffic classifier %s", prefix, traffic_classifier)
        rules = traffic_classifier.map_traffic_classifier_2_redirect_rules()

        return self.synthesize_redirect_bgp_route(rules)

    def _advertise_route_or_default(self, route, label, rd,
                                    lb_consistent_hash_order=0):
        if self.attract_traffic:
            self.log.debug("Advertising default route from VRF %d to "
                           "redirection VRF", self.instance_id)

        route_entry = self._route_for_readvertisement(
            route, label, rd, lb_consistent_hash_order,
            do_default=self.attract_traffic
        )
        self._advertise_route(route_entry)

    def _withdraw_route_or_default(self, route, label, rd,
                                   lb_consistent_hash_order=0):
        if self.attract_traffic:
            self.log.debug("Stop advertising default route from VRF to "
                           "redirection VRF")

        route_entry = self._route_for_readvertisement(
            route, label, rd, lb_consistent_hash_order,
            do_default=self.attract_traffic
        )
        self._withdraw_route(route_entry)

    @log_decorator.log
    def _readvertise(self, route):
        nlri = route.nlri

        self.log.debug("Start re-advertising %s from VRF", nlri.cidr.prefix())
        for _, endpoints in self.localport_2_endpoints.items():
            for endpoint in endpoints:
                port_data = self.mac_2_localport_data[endpoint['mac']]
                label = port_data['label']
                lb_consistent_hash_order = port_data[
                    'lb_consistent_hash_order']
                rd = self.endpoint_2_rd[(endpoint['mac'], endpoint['ip'])]
                self.log.debug("Start re-advertising %s from VRF, with label "
                               "%s and route distinguisher %s",
                               nlri, label, rd)
                # need a distinct RD for each route...
                self._advertise_route_or_default(route, label, rd,
                                                 lb_consistent_hash_order)

        if self.attract_traffic:
            flow_route = self._route_for_redirect_prefix(nlri.cidr.prefix())
            self._advertise_route(flow_route)

        self.readvertised.add(route)

    @log_decorator.log
    def _readvertise_stop(self, route):
        nlri = route.nlri

        self.log.debug("Stop re-advertising %s from VRF", nlri.cidr.prefix())
        for _, endpoints in self.localport_2_endpoints.items():
            for endpoint in endpoints:
                port_data = self.mac_2_localport_data[endpoint['mac']]
                label = port_data['label']
                lb_consistent_hash_order = port_data[
                    'lb_consistent_hash_order']
                rd = self.endpoint_2_rd[(endpoint['mac'], endpoint['ip'])]
                self.log.debug("Stop re-advertising %s from VRF, with label %s"
                               "and route distinguisher %s", nlri, label, rd)
                self._withdraw_route_or_default(route, label, rd,
                                                lb_consistent_hash_order)

        if self.attract_traffic:
            flow_route = self._route_for_redirect_prefix(nlri.cidr.prefix())
            self._withdraw_route(flow_route)

        self.readvertised.remove(route)

    def vif_plugged(self, mac_address, ip_address_prefix, localport,
                    advertise_subnet=False, lb_consistent_hash_order=0):
        vpn_instance.VPNInstance.vif_plugged(self, mac_address,
                                             ip_address_prefix,
                                             localport, advertise_subnet,
                                             lb_consistent_hash_order)

        label = self.mac_2_localport_data[mac_address]['label']
        rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
        for route in self.readvertised:
            self.log.debug("Re-advertising %s with this port as next hop",
                           route.nlri)
            self._advertise_route_or_default(route, label, rd,
                                             lb_consistent_hash_order)

            if self.attract_traffic:
                flow_route = self._route_for_redirect_prefix(
                    route.nlri.cidr.prefix())
                self._advertise_route(flow_route)

    def vif_unplugged(self, mac_address, ip_address_prefix,
                      advertise_subnet=False,
                      lb_consistent_hash_order=0):
        label = self.mac_2_localport_data[mac_address]['label']
        lb_consistent_hash_order = (self.mac_2_localport_data[mac_address]
                                    ["lb_consistent_hash_order"])
        rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
        for route in self.readvertised:
            self.log.debug("Stop re-advertising %s with this port as next hop",
                           route.nlri)
            self._withdraw_route_or_default(route, label, rd,
                                            lb_consistent_hash_order)

            if self.attract_traffic and self.has_only_one_endpoint():
                flow_route = self._route_for_redirect_prefix(
                    route.nlri.cidr.prefix())
                self._withdraw_route(flow_route)

        vpn_instance.VPNInstance.vif_unplugged(self, mac_address,
                                               ip_address_prefix,
                                               advertise_subnet,
                                               lb_consistent_hash_order)

    # Callbacks for BGP route updates (TrackerWorker) ########################

    def _route_2_tracked_entry(self, route):
        if isinstance(route.nlri, ipvpn_routes.IPVPN):
            return route.nlri.cidr.prefix()
        elif isinstance(route.nlri, flowspec.Flow):
            return (flowspec.Flow, route.nlri._rules())
        else:
            self.log.error("We should not receive routes of type %s",
                           type(route.nlri))
            return None

    @utils.synchronized
    @log_decorator.log
    def _new_best_route(self, entry, new_route):

        if isinstance(new_route.nlri, flowspec.Flow):
            if len(new_route.ecoms(exa.TrafficRedirect)) == 1:
                traffic_redirect = new_route.ecoms(exa.TrafficRedirect)
                redirect_rt = "%s:%s" % (traffic_redirect[0].asn,
                                         traffic_redirect[0].target)

                self.start_redirect_traffic(redirect_rt, new_route.nlri.rules)
            else:
                self.log.warning("FlowSpec action or multiple traffic redirect"
                                 " actions not supported: %s",
                                 new_route.ecoms())
        else:
            prefix = entry

            if self.readvertise:
                # check if this is a route we need to re-advertise
                self.log.debug("route RTs: %s", new_route.route_targets)
                self.log.debug("readv from RTs: %s", self.readvertise_from_rts)
                if self._to_readvertise(new_route):
                    self.log.debug("Need to re-advertise %s", prefix)
                    self._readvertise(new_route)

            if not self._imported(new_route):
                self.log.debug("No need to setup dataplane for:%s",
                               prefix)
                return

            encaps = self._check_encaps(new_route)
            if not encaps:
                return

            assert len(new_route.nlri.labels.labels) == 1

            lb_consistent_hash_order = 0
            if new_route.ecoms(exa.ConsistentHashSortOrder):
                lb_consistent_hash_order = new_route.ecoms(
                    exa.ConsistentHashSortOrder)[0].order

            self.dataplane.setup_dataplane_for_remote_endpoint(
                prefix, new_route.nexthop,
                new_route.nlri.labels.labels[0], new_route.nlri, encaps,
                lb_consistent_hash_order)

    @utils.synchronized
    @log_decorator.log
    def _best_route_removed(self, entry, old_route, last):

        if isinstance(old_route.nlri, flowspec.Flow):
            if len(old_route.ecoms(exa.TrafficRedirect)) == 1:
                if last:
                    traffic_redirect = old_route.ecoms(
                        exa.TrafficRedirect)
                    redirect_rt = "%s:%s" % (traffic_redirect[0].asn,
                                             traffic_redirect[0].target)

                    self.stop_redirect_traffic(redirect_rt,
                                               old_route.nlri.rules)
            else:
                self.log.warning("FlowSpec action or multiple traffic redirect"
                                 " actions not supported: %s",
                                 old_route.ecoms())
        else:
            prefix = entry

            if self.readvertise and last:
                # check if this is a route we were re-advertising
                if self._to_readvertise(old_route):
                    self.log.debug("Need to stop re-advertising %s", prefix)
                    self._readvertise_stop(old_route)

            if not self._imported(old_route):
                self.log.debug("No need to update dataplane for:%s",
                               prefix)
                return

            if self._skip_route_removal(last):
                self.log.debug("Skipping removal of non-last route because "
                               "dataplane does not want it")
                return

            encaps = self._check_encaps(old_route)
            if not encaps:
                return

            assert len(old_route.nlri.labels.labels) == 1

            lb_consistent_hash_order = 0
            if old_route.ecoms(exa.ConsistentHashSortOrder):
                lb_consistent_hash_order = old_route.ecoms(
                    exa.ConsistentHashSortOrder)[0].order

            self.dataplane.remove_dataplane_for_remote_endpoint(
                prefix, old_route.nexthop,
                old_route.nlri.labels.labels[0], old_route.nlri, encaps,
                lb_consistent_hash_order)

    # Looking glass ###

    def get_lg_map(self):
        return {
            "readvertised": (lg.SUBTREE, self.get_lg_readvertised_routes),
        }

    def get_lg_readvertised_routes(self, path_prefix):
        return [route.get_lg_local_info(path_prefix)
                for route in self.readvertised]



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import context
import testtools
from testtools import matchers

import neutron.db.api as db
from neutron.plugins.ml2 import driver_api as api
from neutron.tests.unit import testlib_api

from neutron_lib import exceptions as exc

from networking_bagpipe.driver import type_route_target

RT_NN_MIN = 100
RT_NN_MAX = 109
RT_NN_RANGES = [(RT_NN_MIN, RT_NN_MAX)]
UPDATED_RT_NN_RANGES = [(RT_NN_MIN + 5, RT_NN_MAX + 5)]


class RouteTargetTypeTest(testlib_api.SqlTestCase):

    def setUp(self):
        super(RouteTargetTypeTest, self).setUp()
        self.driver = type_route_target.RouteTargetTypeDriver()
        self.driver.rt_nn_ranges = RT_NN_RANGES
        self.driver._sync_route_target_allocations()
        self.session = db.get_reader_session()
        self.context = context.get_admin_context()

    def _get_allocation(self, session, segment):
        return (session.query(type_route_target.RouteTargetAllocation).
                filter_by(rt_nn=segment[api.SEGMENTATION_ID]).
                first())

    def test_sync_rt_nn_allocations(self):
        def check_in_ranges(rt_nn_ranges):
            rt_nn_min, rt_nn_max = rt_nn_ranges[0]
            segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                       api.PHYSICAL_NETWORK: None}

            segment[api.SEGMENTATION_ID] = rt_nn_min - 1
            self.assertIsNone(
                self._get_allocation(self.session, segment))
            segment[api.SEGMENTATION_ID] = rt_nn_max + 1
            self.assertIsNone(
                self._get_allocation(self.session, segment))

            segment[api.SEGMENTATION_ID] = rt_nn_min
            self.assertFalse(
                self._get_allocation(self.session, segment).allocated)
            segment[api.SEGMENTATION_ID] = rt_nn_min + 1
            self.assertFalse(
                self._get_allocation(self.session, segment).allocated)
            segment[api.SEGMENTATION_ID] = rt_nn_max - 1
            self.assertFalse(
                self._get_allocation(self.session, segment).allocated)
            segment[api.SEGMENTATION_ID] = rt_nn_max
            self.assertFalse(
                self._get_allocation(self.session, segment).allocated)

        check_in_ranges(RT_NN_RANGES)
        self.driver.rt_nn_ranges = UPDATED_RT_NN_RANGES
        self.driver._sync_route_target_allocations()
        check_in_ranges(UPDATED_RT_NN_RANGES)

    def _test_sync_allocations_and_allocated(self, rt_nn):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: None,
                   api.SEGMENTATION_ID: rt_nn}
        self.driver.reserve_provider_segment(self.context, segment)

        self.driver.rt_nn_ranges = UPDATED_RT_NN_RANGES
        self.driver._sync_route_target_allocations()

        self.assertTrue(
            self._get_allocation(self.session, segment).allocated)

    def test_sync_allocations_and_allocated_in_initial_range(self):
        self._test_sync_allocations_and_allocated(RT_NN_MIN + 2)

    def test_sync_allocations_and_allocated_in_final_range(self):
        self._test_sync_allocations_and_allocated(RT_NN_MAX + 2)

    def test_partial_segment_is_partial_segment(self):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: None,
                   api.SEGMENTATION_ID: None}
        self.assertTrue(self.driver.is_partial_segment(segment))

    def test_validate_provider_segment(self):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: 'test_net1',
                   api.SEGMENTATION_ID: None}

        with testtools.ExpectedException(exc.InvalidInput):
            self.driver.validate_provider_segment(segment)

        segment[api.PHYSICAL_NETWORK] = None
        self.driver.validate_provider_segment(segment)

        segment[api.SEGMENTATION_ID] = 1
        self.driver.validate_provider_segment(segment)

    def test_reserve_provider_segment(self):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: None,
                   api.SEGMENTATION_ID: 201}
        alloc = self._get_allocation(self.session, segment)
        self.assertIsNone(alloc)
        observed = self.driver.reserve_provider_segment(self.context, segment)
        alloc = self._get_allocation(self.session, observed)
        self.assertTrue(alloc.allocated)

    def test_reserve_provider_segment_already_allocated(self):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: None,
                   api.SEGMENTATION_ID: 201}
        observed = self.driver.reserve_provider_segment(self.context, segment)
        self.assertRaises(type_route_target.RouteTargetInUse,
                          self.driver.reserve_provider_segment,
                          self.context,
                          observed)

    def test_allocate_tenant_segment(self):
        for __ in range(RT_NN_MIN, RT_NN_MAX + 1):
            segment = self.driver.allocate_tenant_segment(self.context)
            alloc = self._get_allocation(self.session, segment)
            self.assertTrue(alloc.allocated)
            rt_nn = segment[api.SEGMENTATION_ID]
            self.assertThat(rt_nn, matchers.GreaterThan(RT_NN_MIN - 1))
            self.assertThat(rt_nn, matchers.LessThan(RT_NN_MAX + 1))

    def test_allocate_tenant_segment_not_available(self):
        for __ in range(RT_NN_MIN, RT_NN_MAX + 1):
            self.driver.allocate_tenant_segment(self.context)
        segment = self.driver.allocate_tenant_segment(self.context)
        self.assertIsNone(segment)

    def test_release_segment(self):
        segment = self.driver.allocate_tenant_segment(self.context)
        self.driver.release_segment(self.context, segment)
        alloc = self._get_allocation(self.session, segment)
        self.assertFalse(alloc.allocated)

    def test_release_segment_unallocated(self):
        segment = {api.NETWORK_TYPE: type_route_target.TYPE_ROUTE_TARGET,
                   api.PHYSICAL_NETWORK: None,
                   api.SEGMENTATION_ID: 201}
        with mock.patch.object(type_route_target.LOG, 'warning') as log_warn:
            self.driver.release_segment(self.context, segment)
            log_warn.assert_called_once_with(
                "Route Target number %(rt_nn)s not found",
                {'rt_nn': 201})



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections
import copy
import socket
import threading

import netaddr
from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import flowspec
from networking_bagpipe.bagpipe_bgp.engine import tracker_worker


LOG = logging.getLogger(__name__)


class TrafficClassifier(object):

    def __init__(self, source_prefix=None, destination_prefix=None,
                 source_port=None, destination_port=None, protocol='tcp'):
        self.source_pfx = source_prefix
        self.destination_pfx = destination_prefix

        if source_port:
            if ":" in source_port:
                self.source_port = tuple(
                    [int(p) for p in source_port.split(":")]
                )
            else:
                self.source_port = int(source_port)
        else:
            self.source_port = source_port

        if destination_port:
            if ":" in destination_port:
                self.destination_port = tuple(
                    [int(p) for p in destination_port.split(":")]
                )
            else:
                self.destination_port = int(destination_port)
        else:
            self.destination_port = destination_port

        self.protocol = protocol

    def __repr__(self):
        return "traffic-classifier:%s" % str(self)

    def __str__(self):
        return "%s-%s,%s-%s,%s" % (self.source_pfx or "*",
                                   self.source_port or "*",
                                   self.destination_pfx or "*",
                                   self.destination_port or "*",
                                   self.protocol)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.__dict__ == other.__dict__)

    def __hash__(self):
        return hash((self.source_pfx, self.source_port,
                     self.destination_pfx, self.destination_port,
                     self.protocol))

    def _interpret_port_rule(self, rule):
        if len(rule) == 1:
            op = rule[0].operations
            port = rule[0].value
            if op == '>=':
                port_range = (int(port), 65535)
            elif op == '<=':
                port_range = (0, int(port))
            elif op == '>':
                port_range = (int(port)+1, 65535)
            elif op == '<':
                port_range = (0, int(port)-1)
            else:
                port_range = int(port)
        else:
            port_range = ()
            for elem in rule:
                op = elem.operations
                port = elem.value
                if op == '>=' or op == '<=':
                    port_range += int(port)
                elif op == '>':
                    port_range += int(port)+1
                elif op == '<':
                    port_range += int(port)-1

        return port_range

    @log_decorator.log
    def _construct_port_rules(self, port_range, flow_object):
        LOG.info("Construct port rules with %s (type %s) on %s object",
                 port_range, type(port_range), flow_object)
        port_rules = list()

        # Check if port range or port number
        if type(port_range) == tuple:
            port_min, port_max = port_range

            if int(port_min) == 0:
                # Operator <
                port_rules.append(flow_object(exa.flow.NumericOperator.LT,
                                              port_max))
            elif int(port_max) == 65535:
                # Operator >
                port_rules.append(flow_object(exa.flow.NumericOperator.GT,
                                              port_min))
            else:
                # Operator >=
                gt_eq_op = (exa.flow.NumericOperator.GT |
                            exa.flow.NumericOperator.EQ)
                # Operator &<=
                lt_eq_op = (exa.flow.NumericOperator.AND |
                            exa.flow.NumericOperator.LT |
                            exa.flow.NumericOperator.EQ)
                port_rules.append(flow_object(gt_eq_op, port_min))
                port_rules.append(flow_object(lt_eq_op, port_max))
        else:
            port_rules.append(flow_object(exa.flow.NumericOperator.EQ,
                                          port_range))

        return port_rules

    def _get_source_prefix(self, rule):
        self.source_pfx = rule[0].cidr.prefix()

    def _get_destination_prefix(self, rule):
        self.destination_pfx = rule[0].cidr.prefix()

    def _parse_any_port(self, rule):
        self.source_port = self._interpret_port_rule(rule)
        self.destination_port = list(self.source_port)

    def _parse_source_port(self, rule):
        self.source_port = self._interpret_port_rule(rule)

    def _parse_destination_port(self, rule):
        self.destination_port = self._interpret_port_rule(rule)

    def _get_protocol(self, rule):
        self.protocol = exa.Protocol.names[rule[0].value]

    def map_traffic_classifier_2_redirect_rules(self):
        rules = list()
        if self.source_pfx:
            ip, mask = self.source_pfx.split('/')
            if netaddr.IPNetwork(self.source_pfx).version == 4:
                rules.append(
                    exa.flow.Flow4Source(socket.inet_pton(socket.AF_INET, ip),
                                         int(mask)))
            elif netaddr.IPNetwork(self.source_pfx).version == 6:
                # TODO(xxx): IPv6 offset ??
                rules.append(
                    exa.flow.Flow6Source(socket.inet_pton(socket.AF_INET6, ip),
                                         int(mask), 0))

        if self.destination_pfx:
            ip, mask = self.destination_pfx.split('/')
            if netaddr.IPNetwork(self.destination_pfx).version == 4:
                rules.append(
                    exa.flow.Flow4Destination(socket.inet_pton(socket.AF_INET,
                                                               ip),
                                              int(mask))
                )
            elif netaddr.IPNetwork(self.destination_pfx).version == 6:
                # TODO(xxx): IPv6 offset ??
                rules.append(
                    exa.flow.Flow6Destination(socket.inet_pton(socket.AF_INET6,
                                                               ip),
                                              int(mask), 0)
                )

        if self.source_port:
            rules += self._construct_port_rules(self.source_port,
                                                exa.flow.FlowSourcePort)

        if self.destination_port:
            rules += self._construct_port_rules(self.destination_port,
                                                exa.flow.FlowDestinationPort)

        if self.protocol:
            rules.append(exa.flow.FlowIPProtocol(exa.flow.NumericOperator.EQ,
                                                 exa.Protocol.named(
                                                     self.protocol)
                                                 ))

        return rules

    def map_redirect_rules_2_traffic_classifier(self, rules):
        components = {1: self._get_destination_prefix,  # FlowDestinationPrefix
                      2: self._get_source_prefix,  # FlowSourcePrefix
                      3: self._get_protocol,  # FlowIPProtocol
                      4: self._parse_any_port,  # FlowAnyPort
                      5: self._parse_destination_port,  # FLowDestinationPort
                      6: self._parse_source_port}  # FlowSourcePort

        for ID, rule in six.iteritems(rules):
            components[ID](rule)


@six.add_metaclass(abc.ABCMeta)
class VPNInstance(tracker_worker.TrackerWorker,
                  threading.Thread,
                  lg.LookingGlassLocalLogger):

    type = None  # set by subclasses: 'ipvpn', 'evpn', etc.
    afi = None
    safi = None

    @log_decorator.log
    def __init__(self, vpn_manager, dataplane_driver,
                 external_instance_id, instance_id, import_rts, export_rts,
                 gateway_ip, mask, readvertise, attract_traffic, fallback=None,
                 **kwargs):

        self.manager = vpn_manager

        self.instance_type = self.__class__.__name__
        self.instance_id = instance_id

        threading.Thread.__init__(self)
        self.setDaemon(True)

        if dataplane_driver.ecmp_support:
            compare_routes = tracker_worker.compare_ecmp
        else:
            compare_routes = tracker_worker.compare_no_ecmp

        tracker_worker.TrackerWorker.__init__(
            self,
            self.manager.bgp_manager, "%s-%d" % (self.instance_type,
                                                 self.instance_id),
            compare_routes)

        lg.LookingGlassLocalLogger.__init__(self,
                                            "%s-%d" % (self.instance_type,
                                                       self.instance_id))
        self.lock = threading.Lock()

        self.import_rts = import_rts
        self.export_rts = export_rts
        self.external_instance_id = external_instance_id
        self.gateway_ip = gateway_ip
        self.mask = mask

        self.fallback = None

        self.afi = self.__class__.afi
        self.safi = self.__class__.safi
        assert isinstance(self.afi, exa.AFI)
        assert isinstance(self.safi, exa.SAFI)

        self.dp_driver = dataplane_driver

        self.instance_label = self.manager.label_allocator.get_new_label(
            "Incoming traffic for %s %d" % (self.instance_type,
                                            self.instance_id))

        self.instance_rd = self.manager.rd_allocator.get_new_rd(
            "Default route distinguisher for %s %d" % (self.instance_type,
                                                       self.instance_id))

        self.localport_data = dict()

        # One local port -> List of endpoints (MAC and IP addresses tuple)
        self.localport_2_endpoints = dict()
        # One endpoint (MAC and IP addresses tuple) -> One route distinguisher
        self.endpoint_2_rd = dict()
        # One MAC address -> One local port
        self.mac_2_localport_data = dict()
        # One IP address ->  Multiple MAC address
        self.ip_address_2_mac = dict()

        # Redirected instances list from which traffic is attracted (based on
        # FlowSpec 5-tuple classification)
        self.redirected_instances = set()

        self.dataplane = self.dp_driver.initialize_dataplane_instance(
            self.instance_id, self.external_instance_id,
            self.gateway_ip, self.mask, self.instance_label, **kwargs)

        for rt in self.import_rts:
            self._subscribe(self.afi, self.safi, rt)
            # Subscribe to FlowSpec routes
            # FIXME(tmorin): this maybe isn't applicable yet to E-VPN yet
            self._subscribe(self.afi, exa.SAFI(exa.SAFI.flow_vpn), rt)

        if readvertise:
            self.readvertise = True
            try:
                self.readvertise_to_rts = readvertise['to_rt']
            except KeyError:
                raise exc.APIException("'readvertise' specified with no "
                                       "'to_rt'")
            self.readvertise_from_rts = readvertise.get('from_rt', [])
            self.log.debug("readvertise enabled, from RT:%s, to %s",
                           self.readvertise_from_rts, self.readvertise_to_rts)
            for rt in self.readvertise_from_rts:
                self._subscribe(self.afi, self.safi, rt)
        else:
            self.log.debug("readvertise not enabled")
            self.readvertise = False

        if self.readvertise and attract_traffic:
            if len(self.readvertise_to_rts) != 1:
                raise exc.APIException("attract_traffic requires exactly one "
                                       "RT to be provided in "
                                       "readvertise/to_rt")
            self.attract_traffic = True
            self.attract_rts = attract_traffic['redirect_rts']
            try:
                self.attract_classifier = attract_traffic['classifier']
            except KeyError:
                raise exc.APIException("'attract_traffic' specified with no "
                                       "'classifier'")
            self.log.debug("Attract traffic enabled with RT: %s and "
                           "classifier: %s", self.attract_rts,
                           self.attract_classifier)
        else:
            self.log.debug("attract traffic not enabled")
            self.attract_traffic = False

        self.dataplane.update_fallback(fallback)

    @utils.synchronized
    def stop(self):
        self._stop()

    @log_decorator.log
    def _stop(self):
        self.dataplane.cleanup()

        self.manager.label_allocator.release(self.instance_label)

        # this makes sure that the thread will be stopped, and any remaining
        # routes/subscriptions are released:
        tracker_worker.TrackerWorker.stop(self)

    @utils.synchronized
    @log_decorator.log
    def stop_if_empty(self):
        self.log.debug("localport_2_endpoints: %s", self.localport_2_endpoints)
        if self.is_empty():
            self._stop()
            return True

        return False

    def is_empty(self):
        return not self.localport_2_endpoints

    def has_enpoint(self, linuxif):
        return self.localport_2_endpoints.get(linuxif) is not None

    def has_only_one_endpoint(self):
        return (len(self.localport_2_endpoints) == 1 and
                len(six.next(six.itervalues(self.localport_2_endpoints))) == 1)

    @log_decorator.log
    def update_route_targets(self, new_import_rts, new_export_rts):
        added_import_rt = set(new_import_rts) - set(self.import_rts)
        removed_import_rt = set(self.import_rts) - set(new_import_rts)

        self.log.debug("%s %d - Added Import RTs: %s",
                       self.instance_type, self.instance_id, added_import_rt)
        self.log.debug("%s %d - Removed Import RTs: %s",
                       self.instance_type, self.instance_id, removed_import_rt)

        # Register to BGP with these route targets
        for rt in added_import_rt:
            self._subscribe(self.afi, self.safi, rt)
            self._subscribe(self.afi, exa.SAFI(exa.SAFI.flow_vpn), rt)

        # Unregister from BGP with these route targets
        for rt in removed_import_rt:
            self._unsubscribe(self.afi, self.safi, rt)
            self._unsubscribe(self.afi, exa.SAFI(exa.SAFI.flow_vpn), rt)

        # Update import and export route targets
        self.import_rts = new_import_rts

        # Re-advertise all routes with new export RTs
        self.log.debug("Exports RTs: %s -> %s", self.export_rts,
                       new_export_rts)
        if frozenset(new_export_rts) != frozenset(self.export_rts):
            self.log.debug("Will re-export routes with new RTs")
            self.export_rts = new_export_rts
            # FIXME: we should only update the routes that
            # are routes of ports plugged to the VPN instance,
            # not all routes which would wrongly include
            # routes that we re-advertise between RTs
            for route_entry in self.get_route_entries():
                self.log.info("Re-advertising route %s with updated RTs (%s)",
                              route_entry.nlri, new_export_rts)

                updated_route_entry = engine.RouteEntry(
                    route_entry.nlri,
                    None,
                    copy.copy(route_entry.attributes))
                # reset the route_targets
                # will RTs originally present in route_entry.attributes
                updated_route_entry.set_route_targets(self.export_rts)

                self.log.debug("   updated route: %s", updated_route_entry)

                self._advertise_route(updated_route_entry)

    def update_fallback(self, fallback):
        if self.fallback != fallback and fallback is not None:
            self.log.info("update fallback: %s", fallback)
            self.fallback = fallback
            self.dataplane.update_fallback(fallback)

    def _parse_ipaddress_prefix(self, ip_address_prefix):
        ip_address = ""
        mask = 0
        try:
            net = netaddr.IPNetwork(ip_address_prefix)
            (ip_address, mask) = (str(net.ip), net.prefixlen)
        except netaddr.core.AddrFormatError:
            raise exc.APIException("Bogus IP prefix: %s" % ip_address_prefix)

        return (ip_address, mask)

    def _gen_encap_extended_communities(self):
        ecommunities = exa.extcoms.ExtendedCommunities()
        for encap in self.dp_driver.supported_encaps():
            if not isinstance(encap, exa.Encapsulation):
                raise Exception("dp_driver.supported_encaps() should "
                                "return a list of Encapsulation objects (%s)",
                                type(encap))

            if encap != exa.Encapsulation(
                    exa.Encapsulation.Type.DEFAULT):
                ecommunities.communities.append(encap)
        # FIXME: si DEFAULT + xxx => adv MPLS
        return ecommunities

    @abc.abstractmethod
    def generate_vif_bgp_route(self, mac_address, ip_prefix, plen, label, rd):
        '''returns a RouteEntry'''
        pass

    def synthesize_vif_bgp_route(self, mac_address, ip_prefix, plen, label,
                                 lb_consistent_hash_order,
                                 route_distinguisher=None):
        rd = route_distinguisher if route_distinguisher else self.instance_rd
        route_entry = self.generate_vif_bgp_route(mac_address, ip_prefix, plen,
                                                  label, rd)
        assert isinstance(route_entry, engine.RouteEntry)

        route_entry.attributes.add(self._gen_encap_extended_communities())
        route_entry.set_route_targets(self.export_rts)

        ecommunities = exa.ExtendedCommunities()
        ecommunities.communities.append(
            exa.ConsistentHashSortOrder(lb_consistent_hash_order))
        route_entry.attributes.add(ecommunities)

        self.log.debug("Synthesized Vif route entry: %s", route_entry)
        return route_entry

    def synthesize_redirect_bgp_route(self, rules):
        self.log.info("synthesize_redirect_bgp_route called for rules %s",
                      rules)
        nlri = flowspec.FlowRouteFactory(self.afi, self.instance_rd)
        for rule in rules:
            nlri.add(rule)

        route_entry = engine.RouteEntry(nlri)

        assert isinstance(route_entry, engine.RouteEntry)

        ecommunities = exa.ExtendedCommunities()

        # checked at __init__:
        assert len(self.readvertise_to_rts) == 1
        rt = self.readvertise_to_rts[0]
        ecommunities.communities.append(
            exa.TrafficRedirect(exa.ASN(int(rt.asn)), int(rt.number))
        )

        route_entry.attributes.add(ecommunities)
        route_entry.set_route_targets(self.attract_rts)

        self.log.debug("Synthesized redirect route entry: %s", route_entry)
        return route_entry

    @utils.synchronized
    @log_decorator.log_info
    def vif_plugged(self, mac_address, ip_address_prefix, localport,
                    advertise_subnet=False,
                    lb_consistent_hash_order=0):
        linuxif = localport['linuxif']
        # Check if this port has already been plugged
        # - Verify port informations consistency
        if mac_address in self.mac_2_localport_data:
            self.log.debug("MAC address already plugged, checking port "
                           "consistency")
            pdata = self.mac_2_localport_data[mac_address]

            if pdata.get("port_info") != localport:
                raise exc.APIException("Port information is not consistent. "
                                       "MAC address cannot be bound to two "
                                       "different ports. Previous plug for "
                                       "port %s (%s != %s)" %
                                       (linuxif, pdata.get("port_info"),
                                        localport))

        try:
            # Parse address/mask
            (ip_prefix, plen) = self._parse_ipaddress_prefix(ip_address_prefix)

            if not advertise_subnet and plen != 32:
                self.log.debug("Using /32 instead of /%d", plen)
                plen = 32

            # - Verify (MAC address, IP address) tuple consistency
            if ip_address_prefix in self.ip_address_2_mac and plen == 32:
                if mac_address not in self.ip_address_2_mac[ip_address_prefix]:
                    raise exc.APIException("Inconsistent endpoint info: %s "
                                           "already bound to a MAC address "
                                           "different from %s" %
                                           (ip_address_prefix, mac_address))
                else:
                    return

            self.log.debug("Plugging port (%s)", ip_prefix)

            pdata = self.mac_2_localport_data.get(mac_address, dict())
            if not pdata:
                pdata['label'] = self.manager.label_allocator.get_new_label(
                    "Incoming traffic for %s %d, interface %s"
                    ", endpoint %s/%s" %
                    (self.instance_type, self.instance_id,
                     linuxif, mac_address, ip_address_prefix)
                )
                pdata["port_info"] = localport
                pdata["lb_consistent_hash_order"] = lb_consistent_hash_order

            endpoint_rd = self.manager.rd_allocator.get_new_rd(
                "Route distinguisher for %s %d, interface %s, "
                "endpoint %s/%s" % (self.instance_type, self.instance_id,
                                    linuxif, mac_address,
                                    ip_address_prefix)
            )

            rd = self.instance_rd if plen == 32 else endpoint_rd

            # Call driver to setup the dataplane for incoming traffic
            self.dataplane.vif_plugged(mac_address, ip_prefix,
                                       localport, pdata['label'])

            self.log.info("Synthesizing and advertising BGP route for VIF %s "
                          "endpoint (%s, %s/%d)", linuxif,
                          mac_address, ip_prefix, plen)
            route_entry = self.synthesize_vif_bgp_route(
                mac_address, ip_prefix, plen,
                pdata['label'], lb_consistent_hash_order, rd)

            self._advertise_route(route_entry)

            if linuxif not in self.localport_2_endpoints:
                self.localport_2_endpoints[linuxif] = list()

            self.localport_2_endpoints[linuxif].append(
                {'mac': mac_address, 'ip': ip_address_prefix}
            )
            self.endpoint_2_rd[(mac_address, ip_address_prefix)] = endpoint_rd
            self.mac_2_localport_data[mac_address] = pdata

            if ip_address_prefix not in self.ip_address_2_mac:
                self.ip_address_2_mac[ip_address_prefix] = list()

            self.ip_address_2_mac[ip_address_prefix].append(mac_address)

        except Exception as e:
            self.log.error("Error in vif_plugged: %s", e)
            if linuxif in self.localport_2_endpoints:
                endpoint = {'mac': mac_address, 'ip': ip_address_prefix}
                if endpoint in self.localport_2_endpoints[linuxif]:
                    self.localport_2_endpoints[linuxif].remove(endpoint)
                if not self.localport_2_endpoints[linuxif]:
                    del self.localport_2_endpoints[linuxif]
            if mac_address in self.mac_2_localport_data:
                del self.mac_2_localport_data[mac_address]
            if ip_address_prefix in self.ip_address_2_mac:
                try:
                    self.ip_address_2_mac[ip_address_prefix].remove(
                        mac_address)
                except ValueError:
                    # ok, value just had not been added
                    pass

            raise e

        self.log.info("localport_2_endpoints: %s", self.localport_2_endpoints)
        self.log.info("endpoint_2_rd: %s", self.endpoint_2_rd)
        self.log.info("mac_2_localport_data: %s",
                      self.mac_2_localport_data)
        self.log.info("ip_address_2_mac: %s", self.ip_address_2_mac)

    @utils.synchronized
    @log_decorator.log_info
    def vif_unplugged(self, mac_address, ip_address_prefix,
                      advertise_subnet=False, lb_consistent_hash_order=0):
        # NOTE(tmorin): move this as a vif_unplugged_precheck, so that
        # in ipvpn.VRF this is done before readvertised route withdrawal

        # Verify port and endpoint (MAC address, IP address) tuple consistency
        pdata = self.mac_2_localport_data.get(mac_address)
        if (not pdata or
                (ip_address_prefix in self.ip_address_2_mac and
                 mac_address not in self.ip_address_2_mac[ip_address_prefix])
                or
                (mac_address, ip_address_prefix) not in self.endpoint_2_rd):
            self.log.error("vif_unplugged called for endpoint (%s, %s), but "
                           "no consistent informations or was not plugged yet",
                           mac_address, ip_address_prefix)
            raise exc.APIException("Inconsistent endpoint (%s, %s) info "
                                   "or endpoint wasn't plugged yet, "
                                   "cannot unplug" %
                                   (mac_address, ip_address_prefix))

        # Finding label and local port informations
        label = pdata.get('label')
        endpoint_rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
        localport = pdata.get('port_info')
        linuxif = localport['linuxif']
        if not label or not localport:
            self.log.error("vif_unplugged called for endpoint (%s, %s), but "
                           "port data (%s, %s) is incomplete",
                           mac_address, ip_address_prefix, label, localport)
            raise Exception("Inconsistent informations for port, bug ?")

        if linuxif in self.localport_2_endpoints:
            # Parse address/mask
            (ip_prefix, plen) = self._parse_ipaddress_prefix(ip_address_prefix)

            last_endpoint = len(
                self.localport_2_endpoints[linuxif]) <= 1

            if not advertise_subnet and plen != 32:
                self.log.debug("Using /32 instead of /%d", plen)
                plen = 32

            rd = self.instance_rd if plen == 32 else endpoint_rd

            self.log.info("Synthesizing and withdrawing BGP route for VIF %s "
                          "endpoint (%s, %s/%d)", linuxif,
                          mac_address, ip_prefix, plen)

            route_entry = self.synthesize_vif_bgp_route(
                mac_address, ip_prefix, plen,
                label, lb_consistent_hash_order, rd)

            self._withdraw_route(route_entry)

            # Unplug endpoint from data plane
            self.dataplane.vif_unplugged(
                mac_address, ip_prefix, localport, label, last_endpoint)

            # Forget data for this port if last endpoint
            if last_endpoint:
                self.log.info("Last endpoint, freeing label %s", label)
                # Free label to the allocator
                self.manager.label_allocator.release(label)

                del self.localport_2_endpoints[linuxif]
            else:
                self.localport_2_endpoints[linuxif].remove(
                    {'mac': mac_address, 'ip': ip_address_prefix}
                )

            del self.endpoint_2_rd[(mac_address, ip_address_prefix)]
            # Free route distinguisher to the allocator
            self.manager.rd_allocator.release(endpoint_rd)

            if not last_endpoint:
                if not any([endpoint['mac'] == mac_address
                            for endpoint
                            in self.localport_2_endpoints[linuxif]]
                           ):
                    del self.mac_2_localport_data[mac_address]
            else:
                del self.mac_2_localport_data[mac_address]

            self.ip_address_2_mac[ip_address_prefix].remove(mac_address)

            if not self.ip_address_2_mac[ip_address_prefix]:
                del self.ip_address_2_mac[ip_address_prefix]
        else:
            self.log.error("vif_unplugged called for endpoint {%s, %s}, but"
                           " port data is incomplete", mac_address,
                           ip_address_prefix)
            raise Exception("bagpipe-bgp bug, check its logs")

        self.log.info("localport_2_endpoints: %s", self.localport_2_endpoints)
        self.log.info("mac_2_localport_data: %s",
                      self.mac_2_localport_data)
        self.log.info("ip_address_2_mac: %s", self.ip_address_2_mac)

    @utils.synchronized
    def register_redirected_instance(self, instance_id):
        self.redirected_instances.add(instance_id)

    @utils.synchronized
    def unregister_redirected_instance(self, instance_id):
        self.redirected_instances.remove(instance_id)

    @utils.synchronized
    @log_decorator.log
    def stop_if_no_redirected_instance(self):
        self.log.debug("redirected_instances: %s", self.redirected_instances)
        if not self.redirected_instances:
            self._stop()
            return True

        return False

    @log_decorator.log
    def start_redirect_traffic(self, redirect_rt, rules):
        self.log.debug("Start redirecting traffic to VPN instance importing "
                       "route target %s based on rules %s", redirect_rt, rules)
        # Create redirection instance if first FlowSpec route for this
        # redirect route target
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            self.external_instance_id, self.type, redirect_rt
        )
        # Create traffic classifier from FlowSpec rules
        classifier = TrafficClassifier()
        classifier.map_redirect_rules_2_traffic_classifier(rules)

        self.dataplane.add_dataplane_for_traffic_classifier(
            classifier,
            redirect_instance.instance_id)

        if not hasattr(self, 'redirect_rt_2_classifiers'):
            # One redirect route target -> Multiple traffic classifiers (One
            # per prefix)
            # Can be inverted to one traffic classifier -> Multiple redirect
            # route targets
            self.redirect_rt_2_classifiers = collections.defaultdict(set)

        self.redirect_rt_2_classifiers[redirect_rt].add(classifier)

    @log_decorator.log
    def stop_redirect_traffic(self, redirect_rt, rules):
        self.log.debug("Stop redirecting traffic to VPN instance importing "
                       "route target %s based on rules %s", redirect_rt, rules)
        # Create traffic classifier from FlowSpec rule
        classifier = TrafficClassifier()
        classifier.map_redirect_rules_2_traffic_classifier(rules)

        if (redirect_rt in self.redirect_rt_2_classifiers and
                classifier in self.redirect_rt_2_classifiers[redirect_rt]):
            self.redirect_rt_2_classifiers[redirect_rt].remove(classifier)

            if not utils.invert_dict_of_sets(
                    self.redirect_rt_2_classifiers)[classifier]:
                self.dataplane.remove_dataplane_for_traffic_classifier(
                    classifier)

            if not self.redirect_rt_2_classifiers[redirect_rt]:
                self.manager.stop_redirect_to_vpn(
                    self.external_instance_id, self.type, redirect_rt
                )
                del self.redirect_rt_2_classifiers[redirect_rt]
        else:
            self.log.error("stop_redirect_traffic called for redirect route "
                           "target %s and classifier %s, but doesn't exist",
                           redirect_rt, classifier)
            raise Exception("bagpipe-bgp bug, check its logs")

    def _check_encaps(self, route):
        '''check that encaps of a route

        returns a list of encaps supported by both the dataplane driver and the
        advertized route (based on BGP Encapsulation community)

        logs a warning if there is no common encap
        '''
        adv_encaps = None
        try:
            adv_encaps = route.ecoms(exa.Encapsulation)
            self.log.debug("Advertized Encaps: %s", adv_encaps)
        except KeyError:
            self.log.debug("no encap advertized, let's use default")

        if not adv_encaps:
            adv_encaps = [exa.Encapsulation(
                exa.Encapsulation.Type.DEFAULT)]

        good_encaps = set(adv_encaps) & set(
            self.dp_driver.supported_encaps())

        if not good_encaps:
            self.log.warning("No encap supported by dataplane driver for route"
                             " %s, advertized: %s, dataplane supports: {%s}",
                             route, adv_encaps,
                             ", ".join([repr(encap) for encap in
                                        self.dp_driver.supported_encaps()]
                                       ))

        return good_encaps

    def _skip_route_removal(self, last):
        '''check if a route withdraw should be skipped

        returns true if the removal of the route should be skipped, based
        whether or not the route removed is the last one and depending on
        the desired behavior for the dataplane driver
        '''
        return not last and not (self.dp_driver.makebefore4break_support or
                                 self.dp_driver.ecmp_support)

    # Callbacks for BGP route updates (TrackerWorker) ########################

    @utils.synchronized
    @log_decorator.log
    def _new_best_route(self, entry, new_route):
        pass

    @utils.synchronized
    @log_decorator.log
    def _best_route_removed(self, entry, old_route, last):
        pass

    # Looking Glass ####

    def get_lg_map(self):
        return {
            "instance_type":         (lg.VALUE, self.instance_type),
            "external_instance_id":  (lg.VALUE, self.external_instance_id),
            "dataplane":             (lg.DELEGATE, self.dataplane),
            "route_targets":         (lg.SUBITEM, self.get_rts),
            "gateway_ip":            (lg.VALUE, self.gateway_ip),
            "subnet_mask":           (lg.VALUE, self.mask),
            "instance_dataplane_id": (lg.VALUE, self.instance_label),
            "ports":                 (lg.SUBTREE, self.get_lg_local_port_data),
            "readvertise":           (lg.SUBITEM, self.get_lg_readvertise),
            "fallback":              (lg.VALUE, self.fallback)
        }

    def get_lg_local_port_data(self, path_prefix):
        r = {}
        for (port, endpoints) in self.localport_2_endpoints.items():
            eps = []
            for endpoint in endpoints:
                eps.append({
                    'label':
                    self.mac_2_localport_data[endpoint['mac']]['label'],
                    'mac_address': endpoint['mac'],
                    'ip_address': endpoint['ip'],
                    'rd': repr(self.endpoint_2_rd[(endpoint['mac'],
                                                   endpoint['ip'])])
                })

            r[port] = {
                'endpoints': eps
            }
        return r

    def get_rts(self):
        return {
            "import": [repr(rt) for rt in self.import_rts],
            "export": [repr(rt) for rt in self.export_rts]
        }

    def get_lg_readvertise(self):
        r = {}
        if self.readvertise:
            r = {'from': [repr(rt) for rt in self.readvertise_from_rts],
                 'to': [repr(rt) for rt in self.readvertise_to_rts]}

            if self.attract_traffic:
                r['attract_traffic'] = {
                    'redirect_rts': [repr(rt) for rt in self.attract_rts],
                    'classifier': self.attract_classifier
                }
        return r
